import numpy as np

from ... import cg, weights
from ... import examples as pysal_examples
from ...cg.kdtree import RADIUS_EARTH_KM, KDTree
from ...common import ATOL, RTOL, pandas
from ...io import geotable as pdio
from ...io.fileio import FileIO as psopen
from .. import contiguity as c
from .. import distance as d
from .. import raster
from ..util import get_points_array

# All instances should test these four methods, and define their own functional
# tests based on common codepaths/estimated weights use cases.


class DistanceMixin:
    polygon_path = pysal_examples.get_path("columbus.shp")
    arc_path = pysal_examples.get_path("stl_hom.shp")
    points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
    euclidean_kdt = KDTree(points, distance_metric="euclidean")

    polygon_f = psopen(polygon_path)  # our file handler
    poly_centroids = get_points_array(polygon_f)  # our iterable
    polygon_f.seek(0)  # go back to head of file

    arc_f = psopen(arc_path)
    arc_points = get_points_array(arc_f)
    arc_f.seek(0)
    arc_kdt = KDTree(
        arc_points, distance_metric="Arc", radius=cg.sphere.RADIUS_EARTH_KM
    )

    cls = object  # class constructor
    known_wi = None  # index of known w entry to compare
    known_w = dict()  # actual w entry
    known_name = known_wi

    def setup_method(self):
        self.__dict__.update(
            {
                k: v
                for k, v in list(DistanceMixin.__dict__.items())
                if not k.startswith("_")
            }
        )

    def test_init(self):
        # test vanilla, named
        raise NotImplementedError(
            "You need to implement this test " "before this module will pass"
        )

    def test_from_shapefile(self):
        # test vanilla, named, sparse
        raise NotImplementedError(
            "You need to implement this test " "before this module will pass"
        )

    def test_from_array(self):
        # test named, sparse
        raise NotImplementedError(
            "You need to implement this test " "before this module will pass"
        )

    def test_from_dataframe(self):
        # test named, columnar, defau
        raise NotImplementedError(
            "You need to implement this test " "before this module will pass"
        )


class TestKNN(DistanceMixin):
    def setup_method(self):
        DistanceMixin.setup_method(self)

        self.known_wi0 = 7
        self.known_w0 = [3, 6, 12, 11]
        self.known_wi1 = 0
        self.known_w1 = [2, 1, 3, 7]

        self.known_wi2 = 4
        self.known_w2 = [1, 3, 9, 12]
        self.known_wi3 = 40
        self.known_w3 = [31, 38, 45, 49]

    ##########################
    # Classmethod tests      #
    ##########################

    def test_init(self):
        w = d.KNN(self.euclidean_kdt, k=2)
        assert w.neighbors[0] == [1, 3]

    def test_from_dataframe(self):
        df = pdio.read_files(self.polygon_path)
        w = d.KNN.from_dataframe(df, k=4)
        assert w.neighbors[self.known_wi0] == self.known_w0
        assert w.neighbors[self.known_wi1] == self.known_w1

        # named geometry
        df.rename(columns={"geometry": "the_geom"}, inplace=True)
        w = d.KNN.from_dataframe(df, k=4, geom_col="the_geom")
        assert w.neighbors[self.known_wi0] == self.known_w0
        assert w.neighbors[self.known_wi1] == self.known_w1

    def test_from_geodataframe(self):
        df = pdio.read_files(self.polygon_path)
        # named active geometry
        df.rename(columns={"geometry": "the_geom"}, inplace=True)
        df = df.set_geometry("the_geom")
        w = d.KNN.from_dataframe(df, k=4)
        assert w.neighbors[self.known_wi0] == self.known_w0
        assert w.neighbors[self.known_wi1] == self.known_w1

    def test_from_array(self):
        w = d.KNN.from_array(self.poly_centroids, k=4)
        assert w.neighbors[self.known_wi0] == self.known_w0
        assert w.neighbors[self.known_wi1] == self.known_w1

    def test_from_shapefile(self):
        w = d.KNN.from_shapefile(self.polygon_path, k=4)
        assert w.neighbors[self.known_wi0] == self.known_w0
        assert w.neighbors[self.known_wi1] == self.known_w1

    ##########################
    # Function/User tests    #
    ##########################

    def test_reweight(self):
        w = d.KNN(self.points, k=2)
        new_point = [(21, 21)]
        wnew = w.reweight(k=4, p=1, new_data=new_point, inplace=False)
        assert wnew[0] == {1: 1.0, 3: 1.0, 4: 1.0, 6: 1.0}

    def test_arcdata(self):
        w = d.KNN.from_shapefile(
            self.polygon_path,
            k=4,
            distance_metric="Arc",
            radius=cg.sphere.RADIUS_EARTH_KM,
        )
        assert w.data.shape[1] == 3


class TestDistanceBand(DistanceMixin):
    def setup_method(self):
        DistanceMixin.setup_method(self)
        self.grid_path = pysal_examples.get_path("lattice10x10.shp")
        self.grid_rook_w = c.Rook.from_shapefile(self.grid_path)
        self.grid_f = psopen(self.grid_path)
        self.grid_points = get_points_array(self.grid_f)
        self.grid_f.seek(0)

        self.grid_kdt = KDTree(self.grid_points)

    ##########################
    # Classmethod tests      #
    ##########################

    def test_init(self):
        w = d.DistanceBand(self.grid_kdt, 1)
        for k, v in w:
            assert v == self.grid_rook_w[k]

    def test_from_shapefile(self):
        w = d.DistanceBand.from_shapefile(self.grid_path, 1)
        for k, v in w:
            assert v == self.grid_rook_w[k]

    def test_from_array(self):
        w = d.DistanceBand.from_array(self.grid_points, 1)
        for k, v in w:
            assert v == self.grid_rook_w[k]

    def test_from_dataframe(self):
        import pandas as pd

        geom_series = pdio.shp.shp2series(self.grid_path)
        random_data = np.random.random(size=len(geom_series))
        df = pd.DataFrame({"obs": random_data, "geometry": geom_series})
        w = d.DistanceBand.from_dataframe(df, 1)
        for k, v in w:
            assert v == self.grid_rook_w[k]

    def test_from_geodataframe(self):
        import geopandas as gpd
        import pandas as pd

        geom_series = pdio.shp.shp2series(self.grid_path)
        random_data = np.random.random(size=len(geom_series))
        df = pd.DataFrame({"obs": random_data, "geometry": geom_series})
        w = d.DistanceBand.from_dataframe(df, 1)
        for k, v in w:
            assert v == self.grid_rook_w[k]
        # named geometry
        df = gpd.GeoDataFrame(df)
        df.rename(columns={"geometry": "the_geom"}, inplace=True)
        w = d.DistanceBand.from_dataframe(df, 1, geom_col="the_geom")
        for k, v in w:
            assert v == self.grid_rook_w[k]

        # named active geometry
        df = df.set_geometry("the_geom")
        w = d.DistanceBand.from_dataframe(df, 1)
        for k, v in w:
            assert v == self.grid_rook_w[k]

    ##########################
    # Function/User tests    #
    ##########################
    def test_integers(self):
        """
        see issue #126
        """
        grid_integers = [tuple(map(int, poly.vertices[0])) for poly in self.grid_f]
        self.grid_f.seek(0)
        grid_dbw = d.DistanceBand(grid_integers, 1)
        for k, v in grid_dbw:
            assert v == self.grid_rook_w[k]

    def test_arcdist(self):
        arc = cg.sphere.arcdist
        kdt = KDTree(
            self.arc_points, distance_metric="Arc", radius=cg.sphere.RADIUS_EARTH_KM
        )
        npoints = self.arc_points.shape[0]
        full = np.array(
            [
                [arc(self.arc_points[i], self.arc_points[j]) for j in range(npoints)]
                for i in range(npoints)
            ]
        )
        maxdist = full.max()
        w = d.DistanceBand(kdt, maxdist, binary=False, alpha=1.0)
        np.testing.assert_allclose(w.sparse.todense(), full)
        assert w.data.shape[1] == 3

    def test_dense(self):
        w_rook = c.Rook.from_shapefile(pysal_examples.get_path("lattice10x10.shp"))
        polys = psopen(pysal_examples.get_path("lattice10x10.shp"))
        centroids = [p.centroid for p in polys]
        w_db = d.DistanceBand(centroids, 1, build_sp=False)

        for k in w_db.id_order:
            np.testing.assert_equal(w_db[k], w_rook[k])

    def test_named(self):
        import pandas as pd

        geom_series = pdio.shp.shp2series(self.grid_path)
        random_data = np.random.random(size=len(geom_series))
        names = [chr(x) for x in range(60, 160)]
        df = pd.DataFrame({"obs": random_data, "geometry": geom_series, "names": names})
        w = d.DistanceBand.from_dataframe(df, 1, ids=df.names)


class TestKernel(DistanceMixin):
    def setup_method(self):
        DistanceMixin.setup_method(self)
        self.known_wi0 = 0
        self.known_w0 = {0: 1, 1: 0.500000049999995, 3: 0.4409830615267465}

        self.known_wi1 = 0
        self.known_w1 = {0: 1.0, 1: 0.33333333333333337, 3: 0.2546440075000701}
        self.known_w1_bw = 15.0

        self.known_wi2 = 0
        self.known_w2 = {
            0: 1.0,
            1: 0.59999999999999998,
            3: 0.55278640450004202,
            4: 0.10557280900008403,
        }
        self.known_w2_bws = [25.0, 15.0, 25.0, 16.0, 14.5, 25.0]

        self.known_wi3 = 0
        self.known_w3 = [1.0, 0.10557289844279438, 9.9999990066379496e-08]
        self.known_w3_abws = [
            [11.180341005532938],
            [11.180341005532938],
            [20.000002000000002],
            [11.180341005532938],
            [14.142137037944515],
            [18.027758180095585],
        ]

        self.known_wi4 = 0
        self.known_w4 = {
            0: 0.3989422804014327,
            1: 0.26741902915776961,
            3: 0.24197074871621341,
        }
        self.known_w4_abws = self.known_w3_abws

        self.known_wi5 = 1
        self.known_w5 = {
            4: 0.0070787731484506233,
            2: 0.2052478782400463,
            3: 0.23051223027663237,
            1: 1.0,
        }

        self.known_wi6 = 0
        self.known_w6 = {0: 1.0, 2: 0.03178906767736345, 1: 9.9999990066379496e-08}
        # stick answers & params here

    ##########################
    # Classmethod tests      #
    ##########################

    def test_init(self):
        w = d.Kernel(self.euclidean_kdt)
        for k, v in list(w[self.known_wi0].items()):
            np.testing.assert_allclose(v, self.known_w0[k], rtol=RTOL)

    def test_from_shapefile(self):
        w = d.Kernel.from_shapefile(self.polygon_path, idVariable="POLYID")
        for k, v in list(w[self.known_wi5].items()):
            np.testing.assert_allclose((k, v), (k, self.known_w5[k]), rtol=RTOL)

        w = d.Kernel.from_shapefile(self.polygon_path, fixed=False)
        for k, v in list(w[self.known_wi6].items()):
            np.testing.assert_allclose((k, v), (k, self.known_w6[k]), rtol=RTOL)

    def test_from_array(self):
        w = d.Kernel.from_array(self.points)
        for k, v in list(w[self.known_wi0].items()):
            np.testing.assert_allclose(v, self.known_w0[k], rtol=RTOL)

    def test_from_dataframe(self):
        df = pdio.read_files(self.polygon_path)
        w = d.Kernel.from_dataframe(df)
        for k, v in list(w[self.known_wi5 - 1].items()):
            np.testing.assert_allclose(v, self.known_w5[k + 1], rtol=RTOL)

    def test_from_geodataframe(self):
        df = pdio.read_files(self.polygon_path)
        # named geometry
        df.rename(columns={"geometry": "the_geom"}, inplace=True)
        w = d.Kernel.from_dataframe(df, geom_col="the_geom")
        for k, v in list(w[self.known_wi5 - 1].items()):
            np.testing.assert_allclose(v, self.known_w5[k + 1], rtol=RTOL)

        # named active geometry
        df = df.set_geometry("the_geom")
        w = d.Kernel.from_dataframe(df)
        for k, v in list(w[self.known_wi5 - 1].items()):
            np.testing.assert_allclose(v, self.known_w5[k + 1], rtol=RTOL)

    ##########################
    # Function/User tests    #
    ##########################

    def test_fixed_bandwidth(self):
        w = d.Kernel(self.points, bandwidth=15.0)
        for k, v in list(w[self.known_wi1].items()):
            np.testing.assert_allclose((k, v), (k, self.known_w1[k]))
        np.testing.assert_allclose(np.ones((w.n, 1)) * 15, w.bandwidth)

        w = d.Kernel(self.points, bandwidth=self.known_w2_bws)
        for k, v in list(w[self.known_wi2].items()):
            np.testing.assert_allclose((k, v), (k, self.known_w2[k]), rtol=RTOL)
        for i in range(w.n):
            np.testing.assert_allclose(w.bandwidth[i], self.known_w2_bws[i], rtol=RTOL)

    def test_adaptive_bandwidth(self):
        w = d.Kernel(self.points, fixed=False)
        np.testing.assert_allclose(
            sorted(w[self.known_wi3].values()), sorted(self.known_w3), rtol=RTOL
        )
        bws = w.bandwidth.tolist()
        np.testing.assert_allclose(bws, self.known_w3_abws, rtol=RTOL)

        w = d.Kernel(self.points, fixed=False, function="gaussian")
        for k, v in list(w[self.known_wi4].items()):
            np.testing.assert_allclose((k, v), (k, self.known_w4[k]), rtol=RTOL)
        bws = w.bandwidth.tolist()
        np.testing.assert_allclose(bws, self.known_w4_abws, rtol=RTOL)

    def test_arcdistance(self):
        w = d.Kernel(
            self.points,
            fixed=True,
            distance_metric="Arc",
            radius=cg.sphere.RADIUS_EARTH_KM,
        )
        assert w.data.shape[1] == 3



# ruff: noqa: F403, F405

import math

import numpy as np
import pytest

from ..shapes import *
from ..standalone import *


class TestBbcommon:
    def test_bbcommon(self):
        b0 = [0, 0, 10, 10]
        b1 = [5, 5, 15, 15]
        assert bbcommon(b0, b1) == 1

    def test_bbcommon_same(self):
        b0 = [0, 0, 10, 10]
        b1 = [0, 0, 10, 10]
        assert bbcommon(b0, b1) == 1

    def test_bbcommon_nested(self):
        b0 = [0, 0, 10, 10]
        b1 = [1, 1, 9, 9]
        assert bbcommon(b0, b1) == 1

    def test_bbcommon_top(self):
        b0 = [0, 0, 10, 10]
        b1 = [3, 5, 6, 15]
        assert bbcommon(b0, b1) == 1

    def test_bbcommon_shared_edge(self):
        b0 = [0, 0, 10, 10]
        b1 = [0, 10, 10, 20]
        assert bbcommon(b0, b1) == 1

    def test_bbcommon_shared_corner(self):
        b0 = [0, 0, 10, 10]
        b1 = [10, 10, 20, 20]
        assert bbcommon(b0, b1) == 1

    def test_bbcommon_floats(self):
        b0 = [0.0, 0.0, 0.1, 0.1]
        b1 = [0.05, 0.05, 0.15, 0.15]
        assert bbcommon(b0, b1) == 1


class TestGetBoundingBox:
    def test_get_bounding_box(self):
        items = [Point((-1, 5)), Rectangle(0, 6, 11, 12)]
        expected = [-1, 5, 11, 12]
        assert expected == get_bounding_box(items)[:]


class TestGetAngleBetween:
    def test_get_angle_between(self):
        ray1 = Ray(Point((0, 0)), Point((1, 0)))
        ray2 = Ray(Point((0, 0)), Point((1, 0)))
        assert get_angle_between(ray1, ray2) == 0.0

    def test_get_angle_between_expect45(self):
        ray1 = Ray(Point((0, 0)), Point((1, 0)))
        ray2 = Ray(Point((0, 0)), Point((1, 1)))
        assert math.degrees(get_angle_between(ray1, ray2)) == 45.0

    def test_get_angle_between_expect90(self):
        ray1 = Ray(Point((0, 0)), Point((1, 0)))
        ray2 = Ray(Point((0, 0)), Point((0, 1)))
        assert math.degrees(get_angle_between(ray1, ray2)) == 90.0


class TestIsCollinear:
    def test_is_collinear(self):
        assert is_collinear(Point((0, 0)), Point((1, 1)), Point((5, 5)))

    def test_is_collinear_expect_false(self):
        assert not is_collinear(Point((0, 0)), Point((1, 1)), Point((5, 0)))

    def test_is_collinear_along_x(self):
        assert is_collinear(Point((0, 0)), Point((1, 0)), Point((5, 0)))

    def test_is_collinear_along_y(self):
        assert is_collinear(Point((0, 0)), Point((0, 1)), Point((0, -1)))

    def test_is_collinear_small_float(self):
        """
        Given: p1 = (0.1, 0.2), p2 = (0.2, 0.3), p3 = (0.3, 0.4)

        Line(p1, p2):  y = mx + b
            m = (0.3-0.2) / (0.2-0.1) = .1/.1 = 1
            y - mx = b
            b = 0.3 - 1*0.2 = 0.1
            b = 0.2 - 1*0.1 = 0.1

            y = 1*x + 0.1

        Line(p2, p3): y = mx + b
            m = (0.4-0.3) / (0.3-0.2) = .1/.1 = 1
            y - mx = b
            b = 0.4 - 1*0.3 = 0.1
            b = 0.4 - 1*0.2 = 0.1

            y = 1*x + 0.1

        Line(p1, p2) == Line(p2 ,p3)

        Therefore ``p1, p2, p3`` are collinear.

        Due to floating point rounding areas the standard test,

            ((p2[0]-p1[0])*(p3[1]-p1[1]) - (p2[1]-p1[1])*(p3[0]-p1[0])) == 0

        will fail. To get around this we use an epsilon. The ``numpy.finfo``
        function return an smallest epsilon for the given data types such that,

            (numpy.finfo(float).eps + 1.0) != 1.0

        Therefore if

            abs(
                (p2[0]-p1[0]) * (p3[1]-p1[1]) - (p2[1]-p1[1]) * (p3[0]-p1[0])
            ) < numpy.finfo(p1[0]).eps

        The points are collinear.

        """

        assert is_collinear(Point((0.1, 0.2)), Point((0.2, 0.3)), Point((0.3, 0.4)))

    def test_is_collinear_random(self):
        for i in range(10):
            a, b, c = np.random.random(3) * 10 ** (i)
            assert is_collinear(Point((a, a)), Point((b, b)), Point((c, c)))

    def test_is_collinear_random2(self):
        for _i in range(1000):
            a, b, c = np.random.random(3)
            assert is_collinear(Point((a, a)), Point((b, b)), Point((c, c)))


class TestGetSegmentsIntersect:
    def test_get_segments_intersect(self):
        seg1 = LineSegment(Point((0, 0)), Point((0, 10)))
        seg2 = LineSegment(Point((-5, 5)), Point((5, 5)))
        assert get_segments_intersect(seg1, seg2)[:] == (0.0, 5.0)

    def test_get_segments_intersect_shared_vert(self):
        seg1 = LineSegment(Point((0, 0)), Point((0, 10)))
        seg2 = LineSegment(Point((-5, 5)), Point((0, 10)))
        assert get_segments_intersect(seg1, seg2)[:] == (0.0, 10.0)

    def test_get_segments_intersect_floats(self):
        seg1 = LineSegment(Point((0, 0)), Point((0, 0.10)))
        seg2 = LineSegment(Point((-0.5, 0.05)), Point((0.5, 0.05)))
        assert get_segments_intersect(seg1, seg2)[:] == (0.0, 0.05)

    def test_get_segments_intersect_angles(self):
        seg1 = LineSegment(Point((0, 0)), Point((1, 1)))
        seg2 = LineSegment(Point((1, 0)), Point((0, 1)))
        assert get_segments_intersect(seg1, seg2)[:] == (0.5, 0.5)

    def test_get_segments_intersect_no_intersect(self):
        seg1 = LineSegment(Point((-5, 5)), Point((5, 5)))
        seg2 = LineSegment(Point((100, 100)), Point((100, 101)))
        assert None is get_segments_intersect(seg1, seg2)

    def test_get_segments_intersect_overlap(self):
        seg1 = LineSegment(Point((0.1, 0.1)), Point((0.6, 0.6)))
        seg2 = LineSegment(Point((0.3, 0.3)), Point((0.9, 0.9)))
        expected = LineSegment(Point((0.3, 0.3)), Point((0.6, 0.6)))
        assert expected == get_segments_intersect(seg1, seg2)

    def test_get_segments_intersect_same(self):
        seg1 = LineSegment(Point((-5, 5)), Point((5, 5)))
        assert seg1 == get_segments_intersect(seg1, seg1)

    def test_get_segments_intersect_nested(self):
        seg1 = LineSegment(Point((0.1, 0.1)), Point((0.9, 0.9)))
        seg2 = LineSegment(Point((0.3, 0.3)), Point((0.6, 0.6)))
        assert seg2 == get_segments_intersect(seg1, seg2)


class TestGetSegmentPointIntersect:
    def test_get_segment_point_intersect(self):
        seg = LineSegment(Point((0, 0)), Point((0, 10)))
        pt = Point((0, 5))
        assert pt == get_segment_point_intersect(seg, pt)

    def test_get_segment_point_intersect_left_end(self):
        seg = LineSegment(Point((0, 0)), Point((0, 10)))
        pt = seg.p1
        assert pt == get_segment_point_intersect(seg, pt)

    def test_get_segment_point_intersect_right_end(self):
        seg = LineSegment(Point((0, 0)), Point((0, 10)))
        pt = seg.p2
        assert pt == get_segment_point_intersect(seg, pt)

    def test_get_segment_point_intersect_angle(self):
        seg = LineSegment(Point((0, 0)), Point((1, 1)))
        pt = Point((0.1, 0.1))
        assert pt == get_segment_point_intersect(seg, pt)

    def test_get_segment_point_intersect_no_intersect(self):
        seg = LineSegment(Point((0, 0)), Point((0, 10)))
        pt = Point((5, 5))
        assert None is get_segment_point_intersect(seg, pt)

    def test_get_segment_point_intersect_no_intersect_collinear(self):
        seg = LineSegment(Point((0, 0)), Point((0, 10)))
        pt = Point((0, 20))
        assert None is get_segment_point_intersect(seg, pt)

    def test_get_segment_point_intersect_floats_1(self):
        seg = LineSegment(Point((0.3, 0.3)), Point((0.9, 0.9)))
        pt = Point((0.5, 0.5))
        assert pt == get_segment_point_intersect(seg, pt)

    def test_get_segment_point_intersect_floats_2(self):
        seg = LineSegment(
            Point((0.0, 0.0)), Point((2.7071067811865475, 2.7071067811865475))
        )
        pt = Point((1.0, 1.0))
        assert pt == get_segment_point_intersect(seg, pt)

    def test_get_segment_point_intersect_floats_no_intersect(self):
        seg = LineSegment(Point((0.3, 0.3)), Point((0.9, 0.9)))
        pt = Point((0.1, 0.1))
        assert None is get_segment_point_intersect(seg, pt)


class TestGetPolygonPointIntersect:
    def test_get_polygon_point_intersect(self):
        poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        pt = Point((0.5, 0.5))
        assert pt == get_polygon_point_intersect(poly, pt)

    def test_get_polygon_point_intersect_on_edge(self):
        poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        pt = Point((1.0, 0.5))
        assert pt == get_polygon_point_intersect(poly, pt)

    def test_get_polygon_point_intersect_on_vertex(self):
        poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        pt = Point((1.0, 1.0))
        assert pt == get_polygon_point_intersect(poly, pt)

    def test_get_polygon_point_intersect_outside(self):
        poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        pt = Point((2.0, 2.0))
        assert None is get_polygon_point_intersect(poly, pt)


class TestGetRectanglePointIntersect:
    def test_get_rectangle_point_intersect(self):
        rect = Rectangle(0, 0, 5, 5)
        pt = Point((1, 1))
        assert pt == get_rectangle_point_intersect(rect, pt)

    def test_get_rectangle_point_intersect_on_edge(self):
        rect = Rectangle(0, 0, 5, 5)
        pt = Point((2.5, 5))
        assert pt == get_rectangle_point_intersect(rect, pt)

    def test_get_rectangle_point_intersect_on_vertex(self):
        rect = Rectangle(0, 0, 5, 5)
        pt = Point((5, 5))
        assert pt == get_rectangle_point_intersect(rect, pt)

    def test_get_rectangle_point_intersect_outside(self):
        rect = Rectangle(0, 0, 5, 5)
        pt = Point((10, 10))
        assert None is get_rectangle_point_intersect(rect, pt)


class TestGetRaySegmentIntersect:
    def test_get_ray_segment_intersect(self):
        ray = Ray(Point((0, 0)), Point((0, 1)))
        seg = LineSegment(Point((-1, 10)), Point((1, 10)))
        assert get_ray_segment_intersect(ray, seg)[:] == (0.0, 10.0)

    def test_get_ray_segment_intersect_orgin(self):
        ray = Ray(Point((0, 0)), Point((0, 1)))
        seg = LineSegment(Point((-1, 0)), Point((1, 0)))
        assert get_ray_segment_intersect(ray, seg)[:] == (0.0, 0.0)

    def test_get_ray_segment_intersect_edge(self):
        ray = Ray(Point((0, 0)), Point((0, 1)))
        seg = LineSegment(Point((0, 2)), Point((2, 2)))
        assert get_ray_segment_intersect(ray, seg)[:] == (0.0, 2.0)

    def test_get_ray_segment_intersect_no_intersect(self):
        ray = Ray(Point((0, 0)), Point((0, 1)))
        seg = LineSegment(Point((10, 10)), Point((10, 11)))
        assert None is get_ray_segment_intersect(ray, seg)

    def test_get_ray_segment_intersect_segment(self):
        ray = Ray(Point((0, 0)), Point((5, 5)))
        seg = LineSegment(Point((1, 1)), Point((2, 2)))
        assert seg == get_ray_segment_intersect(ray, seg)


class TestGetRectangleRectangleIntersection:
    def test_get_rectangle_rectangle_intersection_leftright(self):
        r0 = Rectangle(0, 4, 6, 9)
        r1 = Rectangle(4, 0, 9, 7)
        expected = [4.0, 4.0, 6.0, 7.0]
        assert expected == get_rectangle_rectangle_intersection(r0, r1)[:]

    def test_get_rectangle_rectangle_intersection_topbottom(self):
        r0 = Rectangle(0, 0, 4, 4)
        r1 = Rectangle(2, 1, 6, 3)
        expected = [2.0, 1.0, 4.0, 3.0]
        assert expected == get_rectangle_rectangle_intersection(r0, r1)[:]

    def test_get_rectangle_rectangle_intersection_nested(self):
        r0 = Rectangle(0, 0, 4, 4)
        r1 = Rectangle(2, 1, 3, 2)
        assert r1 == get_rectangle_rectangle_intersection(r0, r1)

    def test_get_rectangle_rectangle_intersection_shared_corner(self):
        r0 = Rectangle(0, 0, 4, 4)
        r1 = Rectangle(4, 4, 8, 8)
        assert Point((4, 4)) == get_rectangle_rectangle_intersection(r0, r1)

    def test_get_rectangle_rectangle_intersection_shared_edge(self):
        r0 = Rectangle(0, 0, 4, 4)
        r1 = Rectangle(0, 4, 4, 8)
        assert LineSegment(
            Point((0, 4)), Point((4, 4))
        ) == get_rectangle_rectangle_intersection(r0, r1)

    def test_get_rectangle_rectangle_intersection_shifted_edge(self):
        r0 = Rectangle(0, 0, 4, 4)
        r1 = Rectangle(2, 4, 6, 8)
        assert LineSegment(
            Point((2, 4)), Point((4, 4))
        ) == get_rectangle_rectangle_intersection(r0, r1)

    def test_get_rectangle_rectangle_intersection_no_intersect(self):
        r0 = Rectangle(0, 0, 4, 4)
        r1 = Rectangle(5, 5, 8, 8)
        assert None is get_rectangle_rectangle_intersection(r0, r1)


class TestGetPolygonPointDist:
    def test_get_polygon_point_dist(self):
        poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        pt = Point((2, 0.5))
        expected = 1.0
        assert expected == get_polygon_point_dist(poly, pt)

    def test_get_polygon_point_dist_inside(self):
        poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        pt = Point((0.5, 0.5))
        expected = 0.0
        assert expected == get_polygon_point_dist(poly, pt)

    def test_get_polygon_point_dist_on_vertex(self):
        poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        pt = Point((1.0, 1.0))
        expected = 0.0
        assert expected == get_polygon_point_dist(poly, pt)

    def test_get_polygon_point_dist_on_edge(self):
        poly = Polygon([Point((0, 0)), Point((1, 0)), Point((1, 1)), Point((0, 1))])
        pt = Point((0.5, 1.0))
        expected = 0.0
        assert expected == get_polygon_point_dist(poly, pt)


class TestGetPointsDist:
    def test_get_points_dist(self):
        pt1 = Point((0.5, 0.5))
        pt2 = Point((0.5, 0.5))
        assert get_points_dist(pt1, pt2) == 0

    def test_get_points_dist_diag(self):
        pt1 = Point((0, 0))
        pt2 = Point((1, 1))
        assert 2 ** (0.5) == get_points_dist(pt1, pt2)

    def test_get_points_dist_along_x(self):
        pt1 = Point((-1000, 1 / 3.0))
        pt2 = Point((1000, 1 / 3.0))
        assert get_points_dist(pt1, pt2) == 2000

    def test_get_points_dist_along_y(self):
        pt1 = Point((1 / 3.0, -500))
        pt2 = Point((1 / 3.0, 500))
        assert get_points_dist(pt1, pt2) == 1000


class TestGetSegmentPointDist:
    def test_get_segment_point_dist(self):
        seg = LineSegment(Point((0, 0)), Point((10, 0)))
        pt = Point((5, 5))
        assert get_segment_point_dist(seg, pt) == (5.0, 0.5)

    def test_get_segment_point_dist_on_end_point(self):
        seg = LineSegment(Point((0, 0)), Point((10, 0)))
        pt = Point((0, 0))
        assert get_segment_point_dist(seg, pt) == (0.0, 0.0)

    def test_get_segment_point_dist_on_middle(self):
        seg = LineSegment(Point((0, 0)), Point((10, 0)))
        pt = Point((5, 0))
        assert get_segment_point_dist(seg, pt) == (0.0, 0.5)

    def test_get_segment_point_diag(self):
        seg = LineSegment(Point((0, 0)), Point((10, 10)))
        pt = Point((5, 5))
        assert pytest.approx(get_segment_point_dist(seg, pt)[0]) == 0.0
        assert pytest.approx(get_segment_point_dist(seg, pt)[1]) == 0.5

    def test_get_segment_point_diag_with_dist(self):
        seg = LineSegment(Point((0, 0)), Point((10, 10)))
        pt = Point((0, 10))
        assert 50 ** (0.5) == pytest.approx(get_segment_point_dist(seg, pt)[0])
        assert pytest.approx(get_segment_point_dist(seg, pt)[1]) == 0.5


class TestGetPointAtAngleAndDist:
    def test_get_point_at_angle_and_dist(self):
        ray = Ray(Point((0, 0)), Point((1, 0)))
        pt = get_point_at_angle_and_dist(ray, math.pi, 1.0)
        assert pytest.approx(pt[0]) == -1.0
        assert pytest.approx(pt[1]) == 0.0

    def test_get_point_at_angle_and_dist_diag(self):
        ray = Ray(Point((0, 0)), Point((1, 1)))
        pt = get_point_at_angle_and_dist(ray, math.pi, 2 ** (0.5))
        assert pytest.approx(pt[0]) == -1.0
        assert pytest.approx(pt[1]) == -1.0

    def test_get_point_at_angle_and_dist_diag_90(self):
        ray = Ray(Point((0, 0)), Point((1, 1)))
        pt = get_point_at_angle_and_dist(ray, -math.pi / 2.0, 2 ** (0.5))
        assert pytest.approx(pt[0]) == 1.0
        assert pytest.approx(pt[1]) == -1.0

    def test_get_point_at_angle_and_dist_diag_45(self):
        ray = Ray(Point((0, 0)), Point((1, 1)))
        pt = get_point_at_angle_and_dist(ray, -math.pi / 4.0, 1)
        assert pytest.approx(pt[0]) == 1.0
        assert pytest.approx(pt[1]) == 0.0


class TestConvexHull:
    def test_convex_hull(self):
        points = [Point((0, 0)), Point((4, 4)), Point((4, 0)), Point((3, 1))]
        assert [Point((0.0, 0.0)), Point((4.0, 0.0)), Point((4.0, 4.0))] == convex_hull(
            points
        )


class TestIsClockwise:
    def test_is_clockwise(self):
        vertices = [Point((0, 0)), Point((0, 10)), Point((10, 0))]
        assert True is is_clockwise(vertices)

    def test_is_clockwise_expect_false(self):
        vertices = [Point((0, 0)), Point((10, 0)), Point((0, 10))]
        assert False is is_clockwise(vertices)

    def test_is_clockwise_big(self):
        vertices = [
            (-106.57798, 35.174143999999998),
            (-106.583412, 35.174141999999996),
            (-106.58417999999999, 35.174143000000001),
            (-106.58377999999999, 35.175542999999998),
            (-106.58287999999999, 35.180543),
            (-106.58263099999999, 35.181455),
            (-106.58257999999999, 35.181643000000001),
            (-106.58198299999999, 35.184615000000001),
            (-106.58148, 35.187242999999995),
            (-106.58127999999999, 35.188243),
            (-106.58138, 35.188243),
            (-106.58108, 35.189442999999997),
            (-106.58104, 35.189644000000001),
            (-106.58028, 35.193442999999995),
            (-106.580029, 35.194541000000001),
            (-106.57974399999999, 35.195785999999998),
            (-106.579475, 35.196961999999999),
            (-106.57922699999999, 35.198042999999998),
            (-106.578397, 35.201665999999996),
            (-106.57827999999999, 35.201642999999997),
            (-106.57737999999999, 35.201642999999997),
            (-106.57697999999999, 35.201543000000001),
            (-106.56436599999999, 35.200311999999997),
            (-106.56058, 35.199942999999998),
            (-106.56048, 35.197342999999996),
            (-106.56048, 35.195842999999996),
            (-106.56048, 35.194342999999996),
            (-106.56048, 35.193142999999999),
            (-106.56048, 35.191873999999999),
            (-106.56048, 35.191742999999995),
            (-106.56048, 35.190242999999995),
            (-106.56037999999999, 35.188642999999999),
            (-106.56037999999999, 35.187242999999995),
            (-106.56037999999999, 35.186842999999996),
            (-106.56037999999999, 35.186552999999996),
            (-106.56037999999999, 35.185842999999998),
            (-106.56037999999999, 35.184443000000002),
            (-106.56037999999999, 35.182943000000002),
            (-106.56037999999999, 35.181342999999998),
            (-106.56037999999999, 35.180433000000001),
            (-106.56037999999999, 35.179943000000002),
            (-106.56037999999999, 35.178542999999998),
            (-106.56037999999999, 35.177790999999999),
            (-106.56037999999999, 35.177143999999998),
            (-106.56037999999999, 35.175643999999998),
            (-106.56037999999999, 35.174444000000001),
            (-106.56037999999999, 35.174043999999995),
            (-106.560526, 35.174043999999995),
            (-106.56478, 35.174043999999995),
            (-106.56627999999999, 35.174143999999998),
            (-106.566541, 35.174144999999996),
            (-106.569023, 35.174157000000001),
            (-106.56917199999999, 35.174157999999998),
            (-106.56938, 35.174143999999998),
            (-106.57061499999999, 35.174143999999998),
            (-106.57097999999999, 35.174143999999998),
            (-106.57679999999999, 35.174143999999998),
            (-106.57798, 35.174143999999998),
        ]
        assert True is is_clockwise(vertices)


class TestPointTouchesRectangle:
    def test_point_touches_rectangle_inside(self):
        rect = Rectangle(0, 0, 10, 10)
        point = Point((5, 5))
        assert point_touches_rectangle(point, rect)

    def test_point_touches_rectangle_on_edge(self):
        rect = Rectangle(0, 0, 10, 10)
        point = Point((10, 5))
        assert point_touches_rectangle(point, rect)

    def test_point_touches_rectangle_on_corner(self):
        rect = Rectangle(0, 0, 10, 10)
        point = Point((10, 10))
        assert point_touches_rectangle(point, rect)

    def test_point_touches_rectangle_outside(self):
        rect = Rectangle(0, 0, 10, 10)
        point = Point((11, 11))
        assert not point_touches_rectangle(point, rect)


class TestGetSharedSegments:
    def test_get_shared_segments(self):
        poly1 = Polygon([Point((0, 0)), Point((0, 1)), Point((1, 1)), Point((1, 0))])
        poly2 = Polygon([Point((1, 0)), Point((1, 1)), Point((2, 1)), Point((2, 0))])
        poly3 = Polygon([Point((0, 1)), Point((0, 2)), Point((1, 2)), Point((1, 1))])
        poly4 = Polygon([Point((1, 1)), Point((1, 2)), Point((2, 2)), Point((2, 1))])
        assert True is get_shared_segments(poly1, poly2, bool_ret=True)
        assert True is get_shared_segments(poly1, poly3, bool_ret=True)
        assert True is get_shared_segments(poly3, poly4, bool_ret=True)
        assert True is get_shared_segments(poly4, poly2, bool_ret=True)

        assert False is get_shared_segments(poly1, poly4, bool_ret=True)
        assert False is get_shared_segments(poly3, poly2, bool_ret=True)

    def test_get_shared_segments_non_bool(self):
        poly1 = Polygon([Point((0, 0)), Point((0, 1)), Point((1, 1)), Point((1, 0))])
        poly2 = Polygon([Point((1, 0)), Point((1, 1)), Point((2, 1)), Point((2, 0))])
        poly3 = Polygon([Point((0, 1)), Point((0, 2)), Point((1, 2)), Point((1, 1))])
        poly4 = Polygon([Point((1, 1)), Point((1, 2)), Point((2, 2)), Point((2, 1))])
        assert (
            LineSegment(Point((1, 0)), Point((1, 1)))
            == get_shared_segments(poly1, poly2)[0]
        )
        assert (
            LineSegment(Point((0, 1)), Point((1, 1)))
            == get_shared_segments(poly1, poly3)[0]
        )
        assert (
            LineSegment(Point((1, 2)), Point((1, 1)))
            == get_shared_segments(poly3, poly4)[0]
        )
        assert (
            LineSegment(Point((2, 1)), Point((1, 1)))
            == get_shared_segments(poly4, poly2)[0]
        )
        # expected =  [LineSegment(Point((1, 1)), Point((1, 0)))]
        # assert expected == get_shared_segments(poly1, poly3)
        # expected =  [LineSegment(Point((1, 1)), Point((1, 0)))]
        # assert expected == get_shared_segments(poly3, poly4)
        # expected =  [LineSegment(Point((1, 1)), Point((1, 0)))]
        # assert expected == get_shared_segments(poly4, poly2)


class TestDistanceMatrix:
    def test_distance_matrix(self):
        points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]
        dist = distance_matrix(np.array(points), 2)
        for i in range(0, len(points)):
            for j in range(i, len(points)):
                x, y = points[i]
                X, Y = points[j]
                d = ((x - X) ** 2 + (y - Y) ** 2) ** (0.5)
                assert dist[i, j] == d



"""
A module for computational geometry.
"""
from .alpha_shapes import *
from .kdtree import *
from .locators import *
from .rtree import *
from .shapes import *
from .sphere import *
from .standalone import *
from .voronoi import *



from . import _accessors as _a
from . import _shapely as _s

# prefer access to shapely computation
_all = {}
_all.update(_s.__dict__)
_all.update(_a.__dict__)

globals().update({_k: _v for _k, _v in list(_all.items()) if not _k.startswith("_")})
_preferred = _a



import os
import tempfile

import libpysal


class TestWIO:
    def setup_method(self):
        self.swmFile1 = libpysal.examples.get_path("ohio.swm")
        self.swmFile2 = libpysal.examples.get_path("us48_CONTIGUITY_EDGES_ONLY.swm")
        self.swmFile3 = libpysal.examples.get_path("us48_INVERSE_DISTANCE.swm")
        self.files = [self.swmFile1, self.swmFile2, self.swmFile3]

    def test_swmio(self):
        for file in self.files:
            f1 = libpysal.io.open(file)
            w1 = f1.read()

            f = tempfile.NamedTemporaryFile(suffix=".swm")
            fname = f.name
            f.close()

            f2 = libpysal.io.open(fname, "w")
            f2.varName = f1.varName
            f2.srs = f1.srs

            f2.write(w1)
            f2.close()
            w2 = libpysal.io.open(fname, "r").read()
            assert w1.pct_nonzero == w2.pct_nonzero
            os.remove(fname)


from .weights import *  # noqa I001
from .distance import *
from .contiguity import *
from .spintW import *
from .util import *
from .user import *
from .set_operations import *
from .spatial_lag import *
from .raster import *
from .gabriel import Gabriel, Delaunay, Relative_Neighborhood



import warnings
from functools import wraps

import numpy
import pandas
from packaging.version import Version
from scipy import sparse, spatial

from libpysal.cg import voronoi_frames

from ._contiguity import _vertex_set_intersection
from ._kernel import _kernel, _kernel_functions, _optimize_bandwidth
from ._utils import (
    _build_coincidence_lookup,
    _induce_cliques,
    _jitter_geoms,
    _validate_geometry_input,
    _vec_euclidean_distances,
)

try:
    from numba import njit  # noqa E401

    HAS_NUMBA = True
except ModuleNotFoundError:
    from libpysal.common import jit as njit

    HAS_NUMBA = False

PANDAS_GE_21 = Version(pandas.__version__) >= Version("2.1.0")

_VALID_GEOMETRY_TYPES = ["Point"]

__author__ = """"
Levi John Wolf (levi.john.wolf@gmail.com)
Martin Fleischmann (martin@martinfleischmann.net)
Serge Rey (sjsrey@gmail.com)
"""


# This is in the module, rather than in `utils`, to ensure that it
# can access `_VALID_GEOMETRY_TYPES` without defining a nested decorator.
def _validate_coincident(triangulator):
    """This is a decorator that validates input for coincident points"""

    @wraps(triangulator)
    def tri_with_validation(
        coordinates,
        ids=None,
        coincident="raise",
        kernel=None,
        bandwidth=None,
        seed=None,
        **kwargs,
    ):
        # validate geometry input
        coordinates, ids, geoms = _validate_geometry_input(
            coordinates, ids=ids, valid_geometry_types=_VALID_GEOMETRY_TYPES
        )

        # check for coincident points
        n_coincident, coincident_lut = _build_coincidence_lookup(geoms)

        # resolve coincident points prior triangulation
        if n_coincident > 0:
            if coincident == "raise":
                raise ValueError(
                    f"There are {len(coincident_lut)} unique locations in "
                    f"the dataset, but {len(geoms)} observations. This means there "
                    "are multiple points in the same location, which is undefined "
                    "for this graph type. To address this issue, consider setting "
                    "`coincident='clique'` or consult the documentation about "
                    "coincident points."
                )
            elif coincident == "jitter":
                coordinates, geoms = _jitter_geoms(coordinates, geoms, seed=seed)
            elif coincident == "clique":
                raise NotImplementedError(
                    "clique-based resolver of coincident points is not yet implemented."
                )
            else:
                raise ValueError(
                    f"Recieved option coincident='{coincident}', but only options "
                    "'raise','clique','jitter' are suppported."
                )

        # generate triangulation (triangulator is the wrapped function)
        heads_ix, tails_ix = triangulator(coordinates, **kwargs)

        # map ids
        heads, tails = ids[heads_ix], ids[tails_ix]

        # process weights
        if kernel is None:
            weights = numpy.ones(heads_ix.shape, dtype=numpy.int8)
        else:
            distances = _vec_euclidean_distances(
                coordinates[heads_ix], coordinates[tails_ix]
            ).squeeze()
            sparse_d = sparse.csc_array((distances, (heads_ix, tails_ix)))
            if bandwidth == "auto":
                bandwidth = _optimize_bandwidth(sparse_d, kernel)
            _, _, weights = _kernel(
                sparse_d,
                metric="precomputed",
                kernel=kernel,
                bandwidth=bandwidth,
                taper=False,
            )
        # create adjacency
        adjtable = pandas.DataFrame.from_dict(
            {"focal": heads, "neighbor": tails, "weight": weights}
        )

        # TODO: fix this
        # reinsert points resolved via clique
        if (n_coincident > 0) & (coincident == "clique"):
            # Note that the kernel is only used to compute a fill value for the clique.
            # In the case of the voronoi weights. Using boxcar with an infinite
            # bandwidth also gives us the correct fill value for the voronoi weight: 1.
            fill_value = _kernel_functions[kernel](numpy.array([0]), bandwidth).item()
            adjtable = _induce_cliques(adjtable, coincident_lut, fill_value=fill_value)

        if PANDAS_GE_21:
            # ensure proper sorting
            sorted_index = (
                adjtable[["focal", "neighbor"]]
                .map(list(ids).index)
                .sort_values(["focal", "neighbor"])
                .index
            )
        else:
            # ensure proper sorting
            sorted_index = (
                adjtable[["focal", "neighbor"]]
                .applymap(list(ids).index)
                .sort_values(["focal", "neighbor"])
                .index
            )

        # return data for Graph.from_arrays
        return heads[sorted_index], tails[sorted_index], weights[sorted_index]

    return tri_with_validation


@_validate_coincident
def _delaunay(coordinates):
    """
    Constructor of the Delaunay graph of a set of input points.
    Relies on scipy.spatial.Delaunay and numba to quickly construct
    a graph from the input set of points. Will be slower without numba,
    and will warn if this is missing.

    Parameters
    ----------
    coordinates :  numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
        geometries containing locations to compute the delaunay triangulation.  If
        a geopandas object with Point geoemtry is provided, the .geometry attribute
        is used. If a numpy.ndarray with shapely geoemtry is used, then the
        coordinates are extracted and used.  If a numpy.ndarray of a shape (2,n) is
        used, it is assumed to contain x, y coordinates.
    ids : numpy.narray (default: None)
        ids to use for each sample in coordinates. Generally, construction functions
        that are accessed via Graph.build_kernel() will set this automatically from
        the index of the input. Do not use this argument directly unless you intend
        to set the indices separately from your input data. Otherwise, use
        data.set_index(ids) to ensure ordering is respected. If None, then the index
        from the input coordinates will be used.
    bandwidth : float (default: None)
        distance to use in the kernel computation. Should be on the same scale as
        the input coordinates.
    kernel : string or callable
        kernel function to use in order to weight the output graph. See the kernel()
        function for more details.

    Notes
    -----
    The Delaunay triangulation can result in quite a few non-local links among
    spatial coordinates. For a more useful graph, consider the weights.Voronoi
    constructor or the Gabriel graph.

    The weights.Voronoi class builds a voronoi diagram among the points, clips the
    Voronoi cells, and then constructs an adjacency graph among the clipped cells.
    This graph among the clipped Voronoi cells generally represents the structure
    of local adjacencies better than the "raw" Delaunay graph.

    The weights.gabriel.Gabriel graph constructs a Delaunay graph, but only
    includes the "short" links in the Delaunay graph.

    However, if the unresricted Delaunay triangulation is needed, this class
    will compute it much more quickly than Voronoi(coordinates, clip=None).
    """

    if not HAS_NUMBA:
        warnings.warn(
            "The numba package is used extensively in this module"
            " to accelerate the computation of graphs. Without numba,"
            " these computations may become unduly slow on large data.",
            stacklevel=3,
        )

    edges, _ = _voronoi_edges(coordinates)
    heads_ix, tails_ix = edges.T

    return heads_ix, tails_ix


@_validate_coincident
def _gabriel(coordinates):
    """
    Constructs the Gabriel graph of a set of points. This graph is a subset of
    the Delaunay triangulation where only "short" links are retained. This
    function is also accelerated using numba, and implemented on top of the
    scipy.spatial.Delaunay class.

    For a link (i,j) connecting node i to j in the Delaunay triangulation
    to be retained in the Gabriel graph, it must pass a point set exclusion test:

    1. Construct the circle C_ij containing link (i,j) as its diameter
    2. If any other node k is contained within C_ij, then remove link (i,j)
       from the graph.
    3. Once all links are evaluated, the remaining graph is the Gabriel graph.

    Parameters
    ----------
    coordinates :  numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
        geometries containing locations to compute the delaunay triangulation.  If
        a geopandas object with Point geoemtry is provided, the .geometry attribute
        is used. If a numpy.ndarray with shapely geoemtry is used, then the
        coordinates are extracted and used.  If a numpy.ndarray of a shape (2,n) is
        used, it is assumed to contain x, y coordinates.
    ids : numpy.narray (default: None)
        ids to use for each sample in coordinates. Generally, construction functions
        that are accessed via Graph.build_kernel() will set this automatically from
        the index of the input. Do not use this argument directly unless you intend
        to set the indices separately from your input data. Otherwise, use
        data.set_index(ids) to ensure ordering is respected. If None, then the index
        from the input coordinates will be used.
    bandwidth : float (default: None)
        distance to use in the kernel computation. Should be on the same scale as
        the input coordinates.
    kernel : string or callable
        kernel function to use in order to weight the output graph. See the kernel()
        function for more details.
    """
    if not HAS_NUMBA:
        warnings.warn(
            "The numba package is used extensively in this module"
            " to accelerate the computation of graphs. Without numba,"
            " these computations may become unduly slow on large data.",
            stacklevel=3,
        )

    edges, dt = _voronoi_edges(coordinates)
    droplist = _filter_gabriel(
        edges,
        dt.points,
    )
    heads_ix, tails_ix = numpy.row_stack(
        list(set(map(tuple, edges)).difference(set(droplist)))
    ).T

    return heads_ix, tails_ix


@_validate_coincident
def _relative_neighborhood(coordinates):
    """
    Constructs the Relative Neighborhood graph from a set of points.
    This graph is a subset of the Delaunay triangulation, where only
    "relative neighbors" are retained. Further, it is a superset of
    the Minimum Spanning Tree, with additional "relative neighbors"
    introduced.

    A relative neighbor pair of points i,j must be closer than the
    maximum distance between i (or j) and each other point k.
    This means that the points are at least as close to one another
    as they are to any other point.

    Parameters
    ----------
    coordinates :  numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
        geometries containing locations to compute the delaunay triangulation.  If
        a geopandas object with Point geoemtry is provided, the .geometry attribute
        is used. If a numpy.ndarray with shapely geoemtry is used, then the
        coordinates are extracted and used.  If a numpy.ndarray of a shape (2,n) is
        used, it is assumed to contain x, y coordinates.
    ids : numpy.narray (default: None)
        ids to use for each sample in coordinates. Generally, construction functions
        that are accessed via Graph.build_kernel() will set this automatically from
        the index of the input. Do not use this argument directly unless you intend
        to set the indices separately from your input data. Otherwise, use
        data.set_index(ids) to ensure ordering is respected. If None, then the index
        from the input coordinates will be used.
    bandwidth : float (default: None)
        distance to use in the kernel computation. Should be on the same scale as
        the input coordinates.
    kernel : string or callable
        kernel function to use in order to weight the output graph. See the kernel()
        function for more details.
    """
    if not HAS_NUMBA:
        warnings.warn(
            "The numba package is used extensively in this module"
            " to accelerate the computation of graphs. Without numba,"
            " these computations may become unduly slow on large data.",
            stacklevel=3,
        )

    edges, dt = _voronoi_edges(coordinates)
    output, _ = _filter_relativehood(edges, dt.points, return_dkmax=False)

    heads_ix, tails_ix, distance = zip(*output, strict=True)
    heads_ix, tails_ix = numpy.asarray(heads_ix), numpy.asarray(tails_ix)

    return heads_ix, tails_ix


@_validate_coincident
def _voronoi(coordinates, clip="extent", rook=True):
    """
    Compute contiguity weights according to a clipped
    Voronoi diagram.

    Parameters
    ---------
    coordinates :  numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
        geometries containing locations to compute the delaunay triangulation.  If
        a geopandas object with Point geoemtry is provided, the .geometry attribute
        is used. If a numpy.ndarray with shapely geoemtry is used, then the
        coordinates are extracted and used.  If a numpy.ndarray of a shape (2,n) is
        used, it is assumed to contain x, y coordinates.
    ids : numpy.narray (default: None)
        ids to use for each sample in coordinates. Generally, construction functions
        that are accessed via Graph.build_kernel() will set this automatically from
        the index of the input. Do not use this argument directly unless you intend
        to set the indices separately from your input data. Otherwise, use
        data.set_index(ids) to ensure ordering is respected. If None, then the index
    clip : str (default: 'bbox')
        An overloaded option about how to clip the voronoi cells passed to
        ``libpysal.cg.voronoi_frames()``.
        Default is ``'extent'``. Options are as follows.

        * ``'none'``/``None`` -- No clip is applied. Voronoi cells may be arbitrarily
            larger that the source map. Note that this may lead to cells that are many
            orders of magnitude larger in extent than the original map.
            Not recommended.
        * ``'bbox'``/``'extent'``/``'bounding box'`` -- Clip the voronoi cells to the
            bounding box of the input points.
        * ``'chull``/``'convex hull'`` -- Clip the voronoi cells to the convex hull of
            the input points.
        * ``'ashape'``/``'ahull'`` -- Clip the voronoi cells to the tightest hull that
            contains all points (e.g. the smallest alphashape, using
            ``libpysal.cg.alpha_shape_auto``).
        * Polygon -- Clip to an arbitrary Polygon.
    rook : bool, optional
        Contiguity method. If True, two geometries are considered neighbours if they
        share at least one edge. If False, two geometries are considered neighbours
        if they share at least one vertex. By default True.

    Notes
    -----
    In theory, the rook contiguity graph for a Voronoi diagram
    is the delaunay triangulation of the generators of the
    voronoi diagram. Yet, this is *not* the case when voronoi
    cells are clipped to an arbitrary shape, including the
    original bounding box of the input points or anything tighter.
    This can arbitrarily delete links present in the delaunay.
    However, clipped voronoi weights make sense over pure
    delaunay triangulations in many applied contexts and
    generally will remove "long" links in the delaunay graph.
    """
    cells, _ = voronoi_frames(coordinates, clip=clip)
    heads_ix, tails_ix, weights = _vertex_set_intersection(cells, rook=rook)

    return heads_ix, tails_ix


#### utilities


@njit
def _edges_from_simplices(simplices):
    """
    Construct the sets of links that correspond to the edges of each
    simplex. Each simplex has three "sides," and thus six undirected
    edges. Thus, the input should be a list of three-length tuples,
    that are then converted into the six non-directed edges for
    each simplex.
    """
    edges = []
    for simplex in simplices:
        edges.append((simplex[0], simplex[1]))
        edges.append((simplex[1], simplex[0]))
        edges.append((simplex[1], simplex[2]))
        edges.append((simplex[2], simplex[1]))
        edges.append((simplex[2], simplex[0]))
        edges.append((simplex[0], simplex[2]))

    return numpy.asarray(edges)


@njit
def _filter_gabriel(edges, coordinates):
    """
    For an input set of edges and coordinates, filter the input edges
    depending on the Gabriel rule:

    For each simplex, let i,j be the diameter of the circle defined by
    edge (i,j), and let k be the third point defining the simplex. The
    limiting case for the Gabriel rule is when k is also on the circle
    with diameter (i,j). In this limiting case, then simplex ijk must
    be a right triangle, and dij**2 = djk**2 + dki**2 (by thales theorem).

    This means that when dij**2 > djk**2 + dki**2, then k is inside the circle.
    In contrast, when dij**2 < djk**2 + dji*2, k is outside of the circle.

    Therefore, it's sufficient to take each observation i, iterate over its
    Delaunay neighbors j,k, and remove links whre dij**2 > djk**2 + dki**2
    in order to construct the Gabriel graph.
    """
    edge_pointer = 0
    n_edges = len(edges)
    to_drop = []
    while edge_pointer < n_edges:
        edge = edges[edge_pointer]
        cardinality = 0
        # look ahead to find all neighbors of edge[0]
        for joff in range(edge_pointer, n_edges):
            next_edge = edges[joff]
            if next_edge[0] != edge[0]:
                break
            cardinality += 1
        for ix in range(edge_pointer, edge_pointer + cardinality):
            i, j = edges[ix]  # lookahead ensures that i is always edge[0]
            dij2 = ((coordinates[i] - coordinates[j]) ** 2).sum()
            for ix2 in range(edge_pointer, edge_pointer + cardinality):
                _, k = edges[ix2]
                if j == k:
                    continue
                dik2 = ((coordinates[i] - coordinates[k]) ** 2).sum()
                djk2 = ((coordinates[j] - coordinates[k]) ** 2).sum()

                if dij2 > (dik2 + djk2):
                    to_drop.append((i, j))
                    to_drop.append((j, i))
        edge_pointer += cardinality
    return set(to_drop)


@njit
def _filter_relativehood(edges, coordinates, return_dkmax=False):
    """
    This is a direct implementation of the algorithm from Toussaint (1980), RNG-2

    1. Compute the delaunay
    2. for each edge of the delaunay (i,j), compute
       dkmax = max(d(k,i), d(k,j)) for k in 1..n, k != i, j
    3. for each edge of the delaunay (i,j), prune
       if any dkmax is greater than d(i,j)
    """
    n = edges.max()
    out = []
    r = []
    for edge in edges:
        i, j = edge
        pi = coordinates[i]
        pj = coordinates[j]
        dkmax = 0
        dij = ((pi - pj) ** 2).sum() ** 0.5
        prune = False
        for k in range(n):
            pk = coordinates[k]
            dik = ((pi - pk) ** 2).sum() ** 0.5
            djk = ((pj - pk) ** 2).sum() ** 0.5
            distances = numpy.array([dik, djk, dkmax])
            dkmax = distances.max()
            prune = dkmax < dij
            if (not return_dkmax) & prune:
                break
        if prune:
            continue
        out.append((i, j, dij))
        if return_dkmax:
            r.append(dkmax)

    return out, r


def _voronoi_edges(coordinates):
    dt = spatial.Delaunay(coordinates)
    edges = _edges_from_simplices(dt.simplices)
    edges = (
        pandas.DataFrame(numpy.asarray(list(edges)))
        .sort_values([0, 1])
        .drop_duplicates()
        .values
    )
    return edges, dt



import os

import geopandas as gpd

from ... import examples as pysal_examples
from ...io.fileio import FileIO as ps_open
from .._contW_lists import QUEEN, ROOK, ContiguityWeightsLists
from ..weights import W


class TestContiguityWeights:
    def setup_method(self):
        """Setup the binning contiguity weights"""
        shpObj = ps_open(pysal_examples.get_path("virginia.shp"), "r")
        self.binningW = ContiguityWeightsLists(shpObj, QUEEN)
        shpObj.close()

    def test_w_type(self):
        assert isinstance(self.binningW, ContiguityWeightsLists)

    def test_queen(self):
        assert QUEEN == 1

    def test_rook(self):
        assert ROOK == 2

    def test_contiguity_weights_lists(self):
        assert hasattr(self.binningW, "w")
        assert issubclass(dict, type(self.binningW.w))
        assert len(self.binningW.w) == 136

    def test_nested_polygons(self):
        # load queen gal file created using Open Geoda.
        geodaW = ps_open(pysal_examples.get_path("virginia.gal"), "r").read()
        # build matching W with pysal
        pysalWb = self.build_W(
            pysal_examples.get_path("virginia.shp"), QUEEN, "POLY_ID"
        )
        # compare output.
        for key in geodaW.neighbors:
            geoda_neighbors = list(map(int, geodaW.neighbors[key]))
            pysalb_neighbors = pysalWb.neighbors[int(key)]
            geoda_neighbors.sort()
            pysalb_neighbors.sort()
            assert geoda_neighbors == pysalb_neighbors

    def test_true_rook(self):
        # load queen gal file created using Open Geoda.
        geodaW = ps_open(pysal_examples.get_path("rook31.gal"), "r").read()
        # build matching W with pysal
        # pysalW = pysal.rook_from_shapefile(pysal_examples.get_path('rook31.shp'),','POLY_ID')
        pysalWb = self.build_W(pysal_examples.get_path("rook31.shp"), ROOK, "POLY_ID")
        # compare output.
        for key in geodaW.neighbors:
            geoda_neighbors = list(map(int, geodaW.neighbors[key]))
            pysalb_neighbors = pysalWb.neighbors[int(key)]
            geoda_neighbors.sort()
            pysalb_neighbors.sort()
            assert geoda_neighbors == pysalb_neighbors

    def test_true_rook2(self):
        # load queen gal file created using Open Geoda.

        stl = pysal_examples.load_example("stl")
        gal_file = test_file = stl.get_path("stl_hom_rook.gal")
        geodaW = ps_open(gal_file, "r").read()
        # build matching W with pysal
        pysalWb = self.build_W(stl.get_path("stl_hom.shp"), ROOK, "POLY_ID_OG")
        # compare output.
        for key in geodaW.neighbors:
            geoda_neighbors = list(map(int, geodaW.neighbors[key]))
            pysalb_neighbors = pysalWb.neighbors[int(key)]
            geoda_neighbors.sort()
            pysalb_neighbors.sort()
            assert geoda_neighbors == pysalb_neighbors

    def test_true_rook3(self):
        # load queen gal file created using Open Geoda.
        geodaW = ps_open(pysal_examples.get_path("virginia_rook.gal"), "r").read()
        # build matching W with pysal
        pysalWb = self.build_W(pysal_examples.get_path("virginia.shp"), ROOK, "POLY_ID")
        # compare output.
        for key in geodaW.neighbors:
            geoda_neighbors = list(map(int, geodaW.neighbors[key]))
            pysalb_neighbors = pysalWb.neighbors[int(key)]
            geoda_neighbors.sort()
            pysalb_neighbors.sort()
            assert geoda_neighbors == pysalb_neighbors

    def test_shapely(self):
        pysalneighbs = ContiguityWeightsLists(
            ps_open(pysal_examples.get_path("virginia.shp")), ROOK
        )
        gdf = gpd.read_file(pysal_examples.get_path("virginia.shp"))
        shplyneighbs = ContiguityWeightsLists(gdf.geometry.tolist(), ROOK)
        assert pysalneighbs.w == shplyneighbs.w
        pysalneighbs = ContiguityWeightsLists(
            ps_open(pysal_examples.get_path("virginia.shp")), QUEEN
        )
        shplyneighbs = ContiguityWeightsLists(gdf.geometry.tolist(), QUEEN)
        assert pysalneighbs.w == shplyneighbs.w

    def build_W(self, shapefile, type, idVariable=None):
        """Building 2 W's the hard way.  We need to do this so we can test both rtree and binning"""
        dbname = os.path.splitext(shapefile)[0] + ".dbf"
        db = ps_open(dbname)
        shpObj = ps_open(shapefile)
        neighbor_data = ContiguityWeightsLists(shpObj, type).w
        neighbors = {}
        weights = {}
        if idVariable:
            ids = db.by_col[idVariable]
            assert len(ids) == len(set(ids))
            for key in neighbor_data:
                id = ids[key]
                if id not in neighbors:
                    neighbors[id] = set()
                neighbors[id].update([ids[x] for x in neighbor_data[key]])
            for key in neighbors:
                neighbors[key] = list(neighbors[key])
            binningW = W(neighbors, id_order=ids)
        else:
            neighbors[key] = list(neighbors[key])
            binningW = W(neighbors)
        return binningW



import numpy as np

from ..spatial_lag import lag_categorical, lag_spatial
from ..util import lat2W
from ..weights import W


class Test_spatial_lag:
    def setup_method(self):
        self.neighbors = {"c": ["b"], "b": ["c", "a"], "a": ["b"]}
        self.weights = {"c": [1.0], "b": [1.0, 1.0], "a": [1.0]}
        self.id_order = ["a", "b", "c"]
        self.weights = {"c": [1.0], "b": [1.0, 1.0], "a": [1.0]}
        self.w = W(self.neighbors, self.weights, self.id_order)
        self.y = np.array([0, 1, 2])
        self.wlat = lat2W(3, 3)
        self.ycat = ["a", "b", "a", "b", "c", "b", "c", "b", "c"]
        self.ycat2 = ["a", "c", "c", "d", "b", "a", "d", "d", "c"]
        self.ym = np.vstack((self.ycat, self.ycat2)).T
        self.random_seed = 503

    def test_lag_spatial(self):
        yl = lag_spatial(self.w, self.y)
        np.testing.assert_array_almost_equal(yl, [1.0, 2.0, 1.0])
        self.w.id_order = ["b", "c", "a"]
        y = np.array([1, 2, 0])
        yl = lag_spatial(self.w, y)
        np.testing.assert_array_almost_equal(yl, [2.0, 1.0, 1.0])
        w = lat2W(3, 3)
        y = np.arange(9)
        yl = lag_spatial(w, y)
        ylc = np.array([4.0, 6.0, 6.0, 10.0, 16.0, 14.0, 10.0, 18.0, 12.0])
        np.testing.assert_array_almost_equal(yl, ylc)
        w.transform = "r"
        yl = lag_spatial(w, y)
        ylc = np.array([2.0, 2.0, 3.0, 3.33333333, 4.0, 4.66666667, 5.0, 6.0, 6.0])
        np.testing.assert_array_almost_equal(yl, ylc)

    def test_lag_categorical(self):
        yl = lag_categorical(self.wlat, self.ycat)
        np.random.seed(self.random_seed)
        known = np.array(["b", "a", "b", "c", "b", "c", "b", "c", "b"])
        np.testing.assert_array_equal(yl, known)
        ym_lag = lag_categorical(self.wlat, self.ym)
        known = np.array(
            [
                ["b", "c"],
                ["a", "c"],
                ["b", "c"],
                ["c", "d"],
                ["b", "d"],
                ["c", "c"],
                ["b", "d"],
                ["c", "d"],
                ["b", "d"],
            ]
        )
        np.testing.assert_array_equal(ym_lag, np.asarray(known))



import math
from functools import cached_property

import numpy as np
import pandas as pd
from scipy import sparse

from libpysal.weights import W

from ._contiguity import (
    _block_contiguity,
    _fuzzy_contiguity,
    _queen,
    _rook,
    _vertex_set_intersection,
)
from ._kernel import _distance_band, _kernel
from ._parquet import _read_parquet, _to_parquet
from ._plotting import _explore_graph, _plot
from ._set_ops import SetOpsMixin
from ._spatial_lag import _lag_spatial
from ._triangulation import _delaunay, _gabriel, _relative_neighborhood, _voronoi
from ._utils import _evaluate_index, _neighbor_dict_to_edges, _sparse_to_arrays

ALLOWED_TRANSFORMATIONS = ("O", "B", "R", "D", "V")

# listed alphabetically
__author__ = """"
Martin Fleischmann (martin@martinfleischmann.net)
Eli Knaap (ek@knaaptime.com)
Serge Rey (sjsrey@gmail.com)
Levi John Wolf (levi.john.wolf@gmail.com)
"""


class Graph(SetOpsMixin):
    """Graph class encoding spatial weights matrices

    The :class:`Graph` is currently experimental
    and its API is incomplete and unstable.
    """

    def __init__(self, adjacency, transformation="O"):
        """Weights base class based on adjacency list

        It is recommenced to use one of the ``from_*`` or ``build_*`` constructors
        rather than invoking ``__init__`` directly.

        Each observation needs to be present in the focal,
        at least as a self-loop with a weight 0.

        Parameters
        ----------
        adjacency : pandas.Series
            A MultiIndexed pandas.Series with ``"focal"`` and ``"neigbor"`` levels
            encoding adjacency, and values encoding weights. By convention,
            isolates are encoded as self-loops with a weight 0.
        transformation : str, default "O"
            weights transformation used to produce the table.

            - **O** -- Original
            - **B** -- Binary
            - **R** -- Row-standardization (global sum :math:`=n`)
            - **D** -- Double-standardization (global sum :math:`=1`)
            - **V** -- Variance stabilizing

        """
        if not isinstance(adjacency, pd.Series):
            raise TypeError(
                f"The adjacency table needs to be a pandas.Series. {type(adjacency)}"
            )
        if not adjacency.index.names == ["focal", "neighbor"]:
            raise ValueError(
                "The index of the adjacency table needs to be a MultiIndex named "
                "['focal', 'neighbor']."
            )
        if not adjacency.name == "weight":
            raise ValueError(
                "The adjacency needs to be named 'weight'. "
                f"'{adjacency.name}' was given instead."
            )
        if not pd.api.types.is_numeric_dtype(adjacency):
            raise ValueError(
                "The 'weight' needs to be of a numeric dtype. "
                f"'{adjacency.dtype}' dtype was given instead."
            )
        if adjacency.isna().any():
            raise ValueError("The adjacency table cannot contain missing values.")
        if transformation.upper() not in ALLOWED_TRANSFORMATIONS:
            raise ValueError(
                f"'transformation' needs to be one of {ALLOWED_TRANSFORMATIONS}. "
                f"'{transformation}' was given instead."
            )
        # adjacency always ordered i-->j on both levels
        ids = adjacency.index.get_level_values(0).unique().values
        adjacency = adjacency.reindex(ids, level=0).reindex(ids, level=1)
        self._adjacency = adjacency
        self.transformation = transformation

    def __getitem__(self, item):
        """Easy lookup based on focal index

        Parameters
        ----------
        item : hashable
            hashable represting an index value

        Returns
        -------
        pandas.Series
            subset of the adjacency table for `item`
        """
        if item in self.isolates:
            return pd.Series(
                [],
                index=pd.Index([], name="neighbor"),
                name="weight",
            )
        return self._adjacency.loc[item]

    def copy(self, deep=True):
        """Make a copy of this Graph's adjacency table and transformation

        Parameters
        ----------
        deep : bool, optional
            Make a deep copy of the adjacency table, by default True

        Returns
        -------
        Graph
            libpysal.graph.Graph as a copy of the original
        """
        return Graph(
            self._adjacency.copy(deep=deep), transformation=self.transformation
        )

    @cached_property
    def adjacency(self):
        """Return a copy of the adjacency list

        Returns
        -------
        pandas.Series
            Underlying adjacency list
        """
        return self._adjacency.copy()

    @classmethod
    def from_W(cls, w):  # noqa N802
        """Create an experimental Graph from libpysal.weights.W object

        Parameters
        ----------
        w : libpysal.weights.W

        Returns
        -------
        Graph
            libpysal.graph.Graph from W
        """
        return cls.from_weights_dict(dict(w))

    def to_W(self):  # noqa N802
        """Convert Graph to a libpysal.weights.W object

        Returns
        -------
        libpysal.weights.W
            representation of graph as a weights.W object
        """
        ids, labels = pd.factorize(
            self._adjacency.index.get_level_values("focal"), sort=False
        )
        neighbors = (
            self._adjacency.reset_index(level=1)
            .groupby(ids)
            .apply(
                lambda group: list(
                    group[
                        ~((group.index == group.neighbor) & (group.weight == 0))
                    ].neighbor
                )
            )
        )
        neighbors.index = labels[neighbors.index]
        weights = (
            self._adjacency.reset_index(level=1)
            .groupby(ids)
            .apply(
                lambda group: list(
                    group[
                        ~((group.index == group.neighbor) & (group.weight == 0))
                    ].weight
                )
            )
        )
        weights.index = labels[weights.index]
        return W(neighbors.to_dict(), weights.to_dict(), id_order=labels.tolist())

    @classmethod
    def from_adjacency(
        cls, adjacency, focal_col="focal", neighbor_col="neighbor", weight_col="weight"
    ):
        """Create a Graph from a pandas DataFrame formatted as an adjacency list

        Parameters
        ----------
        adjacency : pandas.DataFrame
            a dataframe formatted as an ajacency list. Should have columns
            "focal", "neighbor", and "weight", or columns that can be mapped
            to these (e.g. origin, destination, cost)
        focal : str, optional
            name of column holding focal/origin index, by default 'focal'
        neighbor : str, optional
            name of column holding neighbor/destination index, by default 'neighbor'
        weight : str, optional
            name of column holding weight values, by default 'weight'

        Returns
        -------
        Graph
            libpysal.graph.Graph
        """
        cols = dict(
            zip(
                [focal_col, neighbor_col, weight_col],
                ["focal_col", "neighbor_col", "weight_col"],
                strict=True,
            )
        )
        for col in cols:
            assert col in adjacency.columns.tolist(), (
                f'"{col}" was given for `{cols[col]}`, but the '
                f"columns available in `adjacency` are:  {adjacency.columns.tolist()}."
            )
        return cls.from_arrays(
            adjacency[focal_col].values,
            adjacency[neighbor_col].values,
            adjacency[weight_col].values,
        )

    @classmethod
    def from_sparse(cls, sparse, ids=None):
        """Convert a ``scipy.sparse`` array to a PySAL ``Graph`` object.

        Parameters
        ----------
        sparse : scipy.sparse array
            sparse representation of a graph
        ids : list-like, default None
            list-like of ids for geometries that is mappable to
            positions from sparse. If None, the positions are used as labels.

        Returns
        -------
        Graph
            libpysal.graph.Graph based on sparse
        """

        return cls.from_arrays(*_sparse_to_arrays(sparse, ids))

    @classmethod
    def from_arrays(cls, focal_ids, neighbor_ids, weight):
        """Generate Graph from arrays of indices and weights of the same length

        The arrays needs to be sorted in a way ensuring that focal_ids.unique() is
        equal to the index of original observations from which the Graph is being built

        Parameters
        ----------
        focal_index : array-like
            focal indices
        neighbor_index : array-like
            neighbor indices
        weight : array-like
            weights

        Returns
        -------
        Graph
            libpysal.graph.Graph based on arrays
        """

        w = cls(
            pd.Series(
                weight,
                name="weight",
                index=pd.MultiIndex.from_arrays(
                    [focal_ids, neighbor_ids], names=["focal", "neighbor"]
                ),
            )
        )

        return w

    @classmethod
    def from_weights_dict(cls, weights_dict):
        """Generate Graph from a dict of dicts

        Parameters
        ----------
        weights_dict : dictionary of dictionaries
            weights dictionary with the ``{focal: {neighbor: weight}}`` structure.

        Returns
        -------
        Graph
            libpysal.graph.Graph based on weights dictionary of dictionaries
        """
        idx = {f: list(neighbors) for f, neighbors in weights_dict.items()}
        data = {f: list(neighbors.values()) for f, neighbors in weights_dict.items()}
        return cls.from_dicts(idx, data)

    @classmethod
    def from_dicts(cls, neighbors, weights=None):
        """Generate Graph from dictionaries of neighbors and weights

        Parameters
        ----------
        neighbors : dict
            dictionary of neighbors with the ``{focal: [neighbor1, neighbor2]}``
            structure
        weights : dict, optional
            dictionary of neighbors with the ``{focal: [weight1, weight2]}``
            structure. If None, assumes binary weights.

        Returns
        -------
        Graph
            libpysal.graph.Graph based on dictionaries
        """
        head, tail, weight = _neighbor_dict_to_edges(neighbors, weights=weights)
        return cls.from_arrays(head, tail, weight)

    @classmethod
    def build_contiguity(cls, geometry, rook=True, by_perimeter=False, strict=False):
        """Generate Graph from geometry based on contiguity

        Contiguity builder assumes that all geometries are forming a coverage, i.e.
        a non-overlapping mesh and neighbouring geometries share only points or
        segments of their exterior boundaries. In practice, ``build_contiguity`` is
        capable of creating a Graph of partially overlapping geometries when
        ``strict=False, by_perimeter=False``, but that would not strictly follow the
        definition of queen or rook contiguity.

        Parameters
        ----------
        geometry : array-like of shapely.Geometry objects
            Could be geopandas.GeoSeries or geopandas.GeoDataFrame, in which case the
            resulting Graph is indexed by the original index. If an array of
            shapely.Geometry objects is passed, Graph will assume a RangeIndex.
        rook : bool, optional
            Contiguity method. If True, two geometries are considered neighbours if
            they share at least one edge. If False, two geometries are considered
            neighbours if they share at least one vertex. By default True
        by_perimeter : bool, optional
            If True, ``weight`` represents the length of the shared boundary between
            adjacent units, by default False. For row-standardized version of perimeter
            weights, use
            ``Graph.build_contiguity(gdf, by_perimeter=True).transform("r")``.
        strict : bool, optional
            Use the strict topological method. If False, the contiguity is determined
            based on shared coordinates or coordinate sequences representing edges.
            This assumes geometry coverage that is topologically correct. This method
            is faster but can miss some relations. If True, the contiguity is
            determined based on geometric relations that do not require precise
            topology. This method is slower but will result in correct contiguity
            even if the topology of geometries is not optimal. By default False.

        Returns
        -------
        Graph
            libpysal.graph.Graph encoding contiguity weights
        """
        ids = _evaluate_index(geometry)

        if hasattr(geometry, "geometry"):
            # potentially cast GeoDataFrame to GeoSeries
            geometry = geometry.geometry

        if strict:
            # use shapely-based constructors
            if rook:
                return cls.from_arrays(
                    *_rook(geometry, ids=ids, by_perimeter=by_perimeter)
                )
            return cls.from_arrays(
                *_queen(geometry, ids=ids, by_perimeter=by_perimeter)
            )

        # use vertex-based constructor
        return cls.from_arrays(
            *_vertex_set_intersection(
                geometry, rook=rook, ids=ids, by_perimeter=by_perimeter
            )
        )

    @classmethod
    def build_kernel(
        cls,
        data,
        kernel="gaussian",
        k=None,
        bandwidth=None,
        metric="euclidean",
        p=2,
        coincident="raise",
    ):
        """Generate Graph from geometry data based on a kernel function

        Parameters
        ----------
        data : numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
            geometries over which to compute a kernel. If a geopandas object with Point
            geoemtry is provided, the .geometry attribute is used. If a numpy.ndarray
            with shapely geoemtry is used, then the coordinates are extracted and used.
            If a numpy.ndarray of a shape (2,n) is used, it is assumed to contain x, y
            coordinates. If metric="precomputed", data is assumed to contain a
            precomputed distance metric.
        kernel : string or callable (default: 'gaussian')
            kernel function to apply over the distance matrix computed by `metric`.
            The following kernels are supported:

            - ``"triangular"``:
            - ``"parabolic"``:
            - ``"gaussian"``:
            - ``"bisquare"``:
            - ``"cosine"``:
            - ``'boxcar'``/discrete: all distances less than `bandwidth` are 1, and all
              other distances are 0
            - ``"identity"``/None : do nothing, weight similarity based on raw distance
            - ``callable`` : a user-defined function that takes the distance vector and
              the bandwidth and returns the kernel: kernel(distances, bandwidth)

        k : int (default: None)
            number of nearest neighbors used to truncate the kernel. This is assumed
            to be constant across samples. If None, no truncation is conduted.
        bandwidth : float (default: None)
            distance to use in the kernel computation. Should be on the same scale as
            the input coordinates.
        metric : string or callable (default: 'euclidean')
            distance function to apply over the input coordinates. Supported options
            depend on whether or not scikit-learn is installed. If so, then any
            distance function supported by scikit-learn is supported here. Otherwise,
            only euclidean, minkowski, and manhattan/cityblock distances are admitted.
        p : int (default: 2)
            parameter for minkowski metric, ignored if metric != "minkowski".
        coincident: str, optional (default "raise")
            Method for handling coincident points when ``k`` is not None. Options are
            ``'raise'`` (raising an exception when coincident points are present),
            ``'jitter'`` (randomly displace coincident points to produce uniqueness), &
            ``'clique'`` (induce fully-connected sub cliques for coincident points).

        Returns
        -------
        Graph
            libpysal.graph.Graph encoding kernel weights
        """
        ids = _evaluate_index(data)

        head, tail, weight = _kernel(
            data,
            bandwidth=bandwidth,
            metric=metric,
            kernel=kernel,
            k=k,
            p=p,
            ids=ids,
            coincident=coincident,
        )

        return cls.from_arrays(head, tail, weight)

    @classmethod
    def build_knn(cls, data, k, metric="euclidean", p=2, coincident="raise"):
        """Generate Graph from geometry data based on k-nearest neighbors search

        Parameters
        ----------
        data : numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
            geometries over which to compute a kernel. If a geopandas object with Point
            geoemtry is provided, the .geometry attribute is used. If a numpy.ndarray
            with shapely geoemtry is used, then the coordinates are extracted and used.
            If a numpy.ndarray of a shape (2,n) is used, it is assumed to contain x, y
            coordinates.
        k : int
            number of nearest neighbors.
        metric : string or callable (default: 'euclidean')
            distance function to apply over the input coordinates. Supported options
            depend on whether or not scikit-learn is installed. If so, then any
            distance function supported by scikit-learn is supported here. Otherwise,
            only euclidean, minkowski, and manhattan/cityblock distances are admitted.
        p : int (default: 2)
            parameter for minkowski metric, ignored if metric != "minkowski".
        coincident: str, optional (default "raise")
            Method for handling coincident points. Options include
            ``'raise'`` (raising an exception when coincident points are present),
            ``'jitter'`` (randomly displace coincident points to produce uniqueness), &
            ``'clique'`` (induce fully-connected sub cliques for coincident points).


        Returns
        -------
        Graph
            libpysal.graph.Graph encoding KNN weights
        """
        ids = _evaluate_index(data)

        head, tail, weight = _kernel(
            data,
            bandwidth=np.inf,
            metric=metric,
            kernel="boxcar",
            k=k,
            p=p,
            ids=ids,
            coincident=coincident,
        )

        return cls.from_arrays(head, tail, weight)

    @classmethod
    def build_triangulation(
        cls,
        data,
        method="delaunay",
        bandwidth=np.inf,
        kernel="boxcar",
        clip="extent",
        rook=True,
        coincident="raise",
    ):
        """Generate Graph from geometry based on triangulation

        Parameters
        ----------
        data : numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
            geometries containing locations to compute the
            delaunay triangulation. If a geopandas object with Point
            geoemtry is provided, the .geometry attribute is used. If a numpy.ndarray
            with shapely geoemtry is used, then the coordinates are extracted and used.
            If a numpy.ndarray of a shape (2,n) is used, it is assumed to contain x, y
            coordinates.
        method : str, (default "delaunay")
            method of extracting the weights from triangulation. Supports:

            - ``"delaunay"``
            - ``"gabriel"``
            - ``"relative_neighborhood"``
            - ``"voronoi"``

        bandwidth : float, optional
            distance to use in the kernel computation. Should be on the same scale as
            the input coordinates, by default numpy.inf
        kernel : str, optional
            kernel function to use in order to weight the output graph. See
            :meth:`Graph.build_kernel` for details. By default "boxcar"
        clip : str (default: 'bbox')
            Clipping method when ``method="voronoi"``. Ignored otherwise.
            Default is ``'extent'``. Options are as follows.

            - ``'none'``/``None``: No clip is applied. Voronoi cells may be
              arbitrarily larger that the source map. Note that this may lead to
              cells that are many orders of magnitude larger in extent than the
              original map. Not recommended.
            - ``'bbox'``/``'extent'``/``'bounding box'``: Clip the voronoi cells to
              the bounding box of the input points.
            - ``'chull``/``'convex hull'``: Clip the voronoi cells to the convex hull
              of the input points.
            - ``'ashape'``/``'ahull'``: Clip the voronoi cells to the tightest hull
              that contains all points (e.g. the smallest alphashape, using
              :func:`libpysal.cg.alpha_shape_auto`).
            - ``shapely.Polygon``: Clip to an arbitrary Polygon.

        rook : bool, optional
            Contiguity method when ``method="voronoi"``. Ignored otherwise.
            If True, two geometries are considered neighbours if they
            share at least one edge. If False, two geometries are considered neighbours
            if they share at least one vertex. By default True
        coincident: str, optional (default "raise")
            Method for handling coincident points. Options include
            ``'raise'`` (raising an exception when coincident points are present),
            ``'jitter'`` (randomly displace coincident points to produce uniqueness), &
            ``'clique'`` (induce fully-connected sub cliques for coincident points).

        Returns
        -------
        Graph
            libpysal.graph.Graph encoding triangulation weights
        """
        ids = _evaluate_index(data)

        if method == "delaunay":
            head, tail, weights = _delaunay(
                data, ids=ids, bandwidth=bandwidth, kernel=kernel, coincident=coincident
            )
        elif method == "gabriel":
            head, tail, weights = _gabriel(
                data, ids=ids, bandwidth=bandwidth, kernel=kernel, coincident=coincident
            )
        elif method == "relative_neighborhood":
            head, tail, weights = _relative_neighborhood(
                data, ids=ids, bandwidth=bandwidth, kernel=kernel, coincident=coincident
            )
        elif method == "voronoi":
            head, tail, weights = _voronoi(
                data, ids=ids, clip=clip, rook=rook, coincident=coincident
            )
        else:
            raise ValueError(
                f"Method '{method}' is not supported. Use one of ['delaunay', "
                "'gabriel', 'relative_neighborhood', 'voronoi']."
            )

        return cls.from_arrays(head, tail, weights)

    @classmethod
    def build_distance_band(
        cls, data, threshold, binary=True, alpha=-1.0, kernel=None, bandwidth=None
    ):
        """Generate Graph from geometry based on a distance band

        Parameters
        ----------
        data : numpy.ndarray, geopandas.GeoSeries, geopandas.GeoDataFrame
            geometries containing locations to compute the
            delaunay triangulation. If a geopandas object with Point
            geometry is provided, the .geometry attribute is used. If a numpy.ndarray
            with shapely geometry is used, then the coordinates are extracted and used.
            If a numpy.ndarray of a shape (2,n) is used, it is assumed to contain x, y
            coordinates.
        threshold : float
            distance band
        binary : bool, optional
            If True :math:`w_{ij}=1` if :math:`d_{i,j}<=threshold`, otherwise
            :math:`w_{i,j}=0`.
            If False :math:`wij=dij^{alpha}`, by default True.
        alpha : float, optional
            distance decay parameter for weight (default -1.0)
            if alpha is positive the weights will not decline with
            distance. Ignored if ``binary=True`` or ``kernel`` is not None.
        kernel : str, optional
            kernel function to use in order to weight the output graph. See
            :meth:`Graph.build_kernel` for details. Ignored if ``binary=True``.
        bandwidth : float (default: None)
            distance to use in the kernel computation. Should be on the same scale as
            the input coordinates. Ignored if ``binary=True`` or ``kernel=None``.

        Returns
        -------
        Graph
            libpysal.graph.Graph encoding distance band weights
        """
        ids = _evaluate_index(data)

        dist = _distance_band(data, threshold)

        if binary:
            head, tail, weight = _kernel(
                dist,
                kernel="boxcar",
                metric="precomputed",
                ids=ids,
                bandwidth=np.inf,
            )
        elif kernel is not None:
            head, tail, weight = _kernel(
                dist,
                kernel=kernel,
                metric="precomputed",
                ids=ids,
                bandwidth=bandwidth,
            )
        else:
            head, tail, weight = _kernel(
                dist,
                kernel=lambda distances, alpha: np.power(distances, alpha),
                metric="precomputed",
                ids=ids,
                bandwidth=alpha,
            )

        adjacency = pd.DataFrame.from_dict(
            {"focal": head, "neighbor": tail, "weight": weight}
        ).set_index("focal")

        # drop diagonal
        counts = adjacency.index.value_counts()
        no_isolates = counts[counts > 1]
        adjacency = adjacency[
            ~(
                adjacency.index.isin(no_isolates.index)
                & (adjacency.index == adjacency.neighbor)
            )
        ]
        # set isolates to 0 - distance band should never contain self-weight
        adjacency.loc[~adjacency.index.isin(no_isolates.index), "weight"] = 0

        return cls.from_arrays(
            adjacency.index.values, adjacency.neighbor.values, adjacency.weight.values
        )

    @classmethod
    def build_block_contiguity(cls, regimes):
        """Generate Graph from block contiguity (regime neighbors)

        Block contiguity structures are relevant when defining neighbor relations
        based on membership in a regime. For example, all counties belonging to
        the same state could be defined as neighbors, in an analysis of all
        counties in the US.

        Parameters
        ----------
        regimes : list-like
            list-like of regimes. If pandas.Series, its index is used to encode Graph.
            Otherwise a default RangeIndex is used.

        Returns
        -------
        Graph
            libpysal.graph.Graph encoding block contiguity
        """
        ids = _evaluate_index(regimes)

        return cls.from_dicts(_block_contiguity(regimes, ids=ids))

    @classmethod
    def build_fuzzy_contiguity(
        cls,
        geometry,
        tolerance=None,
        buffer=None,
        predicate="intersects",
    ):
        """Generate Graph from fuzzy contiguity

        Fuzzy contiguity relaxes the notion of contiguity neighbors
        for the case of geometry collections that violate the condition
        of planar enforcement. It handles three types of conditions present
        in such collections that would result in missing links when using
        the regular contiguity methods.

        The first are edges for nearby polygons that should be shared, but are
        digitized separately for the individual polygons and the resulting edges
        do not coincide, but instead the edges intersect. This case can also be
        covered by ``build_contiguty`` with the ``strict=False`` parameter.

        The second case is similar to the first, only the resultant edges do not
        intersect but are "close". The optional buffering of geometry then closes the
        gaps between the polygons and a resulting intersection is encoded as a link.

        The final case arises when one polygon is "inside" a second polygon but is not
        encoded to represent a hole in the containing polygon.

        It is also possible to create a contiguity based on a custom spatial predicate.

        Parameters
        ----------
        geoms :  array-like of shapely.Geometry objects
            Could be geopandas.GeoSeries or geopandas.GeoDataFrame, in which case the
            resulting Graph is indexed by the original index. If an array of
            shapely.Geometry objects is passed, Graph will assume a RangeIndex.
        tolerance : float, optional
            The percentage of the length of the minimum side of the bounding rectangle
            for the ``geoms`` to use in determining the buffering distance. Either
            ``tolerance`` or ``buffer`` may be specified but not both.
            By default None.
        buffer : float, optional
            Exact buffering distance in the units of ``geoms.crs``. Either
            ``tolerance`` or ``buffer`` may be specified but not both.
            By default None.
        predicate : str, optional
            The predicate to use for determination of neighbors. Default is
            'intersects'. If None is passed, neighbours are determined based
            on the intersection of bounding boxes. See the documentation of
            ``geopandas.GeoSeries.sindex.query`` for allowed predicates.

        Returns
        -------
        Graph
            libpysal.graph.Graph encoding fuzzy contiguity
        """
        ids = _evaluate_index(geometry)

        heads, tails, weights = _fuzzy_contiguity(
            geometry, ids, tolerance=tolerance, buffer=buffer, predicate=predicate
        )

        return cls.from_arrays(heads, tails, weights)

    @cached_property
    def neighbors(self):
        """Get neighbors dictionary

        Notes
        -----
        It is recommended to work directly with :meth:`Graph.adjacency` rather than
        using the :meth:`Graph.neighbors`.

        Returns
        -------
        dict
            dict of tuples representing neighbors
        """
        return (
            self._adjacency.reset_index(level=1)
            .groupby(level=0)
            .apply(
                lambda group: tuple(
                    group[
                        ~((group.index == group.neighbor) & (group.weight == 0))
                    ].neighbor
                )
            )
            .to_dict()
        )

    @cached_property
    def weights(self):
        """Get weights dictionary

        Notes
        -----
        It is recommended to work directly with :meth:`Graph.adjacency` rather than
        using the :meth:`Graph.weights`.

        Returns
        -------
        dict
            dict of tuples representing weights
        """
        return (
            self._adjacency.reset_index(level=1)
            .groupby(level=0)
            .apply(
                lambda group: tuple(
                    group[
                        ~((group.index == group.neighbor) & (group.weight == 0))
                    ].weight
                )
            )
            .to_dict()
        )

    @cached_property
    def sparse(self):
        """Return a scipy.sparse array (COO)

        Returns
        -------
        scipy.sparse.COO
            sparse representation of the adjacency
        """
        # pivot to COO sparse matrix and cast to array
        return sparse.coo_array(
            self._adjacency.astype("Sparse[float]").sparse.to_coo(sort_labels=True)[0]
        )

    def transform(self, transformation):
        """Transformation of weights

        Parameters
        ----------
        transformation : str
            Transformation method. The following are
            valid transformations.

            - **B** -- Binary
            - **R** -- Row-standardization (global sum :math:`=n`)
            - **D** -- Double-standardization (global sum :math:`=1`)
            - **V** -- Variance stabilizing

        Returns
        -------
        Graph
            transformed weights

        Raises
        ------
        ValueError
            Value error for unsupported transformation
        """
        transformation = transformation.upper()

        if self.transformation == transformation:
            return self.copy()

        if transformation == "R":
            standardized = (
                (self._adjacency / self._adjacency.groupby(level=0).transform("sum"))
                .fillna(0)
                .values
            )  # isolate comes as NaN -> 0

        elif transformation == "D":
            standardized = (self._adjacency / self._adjacency.sum()).values

        elif transformation == "B":
            standardized = self._adjacency.astype(bool).astype(int)

        elif transformation == "V":
            s = self._adjacency.groupby(level=0).transform(
                lambda group: group / math.sqrt((group**2).sum())
            )
            n_q = self.n / s.sum()
            standardized = (s * n_q).fillna(0).values  # isolate comes as NaN -> 0

        else:
            raise ValueError(
                f"Transformation '{transformation}' is not supported. "
                f"Use one of {ALLOWED_TRANSFORMATIONS[1:]}"
            )

        standardized_adjacency = pd.Series(
            standardized, name="weight", index=self._adjacency.index
        )
        return Graph(standardized_adjacency, transformation)

    @cached_property
    def _components(self):
        """helper for n_components and component_labels"""
        # TODO: remove casting to matrix once scipy supports arrays here
        return sparse.csgraph.connected_components(sparse.coo_matrix(self.sparse))

    @cached_property
    def n_components(self):
        """Get a number of connected components

        Returns
        -------
        int
            number of components
        """
        return self._components[0]

    @cached_property
    def component_labels(self):
        """Get component labels per observation

        Returns
        -------
        numpy.array
            Array of component labels
        """
        return pd.Series(
            self._components[1], index=self.unique_ids, name="component labels"
        )

    @cached_property
    def cardinalities(self):
        """Number of neighbors for each observation

        Returns
        -------
        pandas.Series
            Series with a number of neighbors per each observation
        """
        cardinalities = self._adjacency.astype(bool).groupby(level=0).sum()
        cardinalities.name = "cardinalities"
        return cardinalities

    @cached_property
    def isolates(self):
        """Index of observations with no neighbors

        Isolates are encoded as a self-loop with
        the weight == 0 in the adjacency table.

        Returns
        -------
        pandas.Index
            Index with a subset of observations that do not have any neighbor
        """
        nulls = self._adjacency[self._adjacency == 0].reset_index(level=1)
        # since not all zeros are necessarily isolates, do the focal == neighbor check
        return nulls[nulls.index == nulls.neighbor].index.unique()

    @cached_property
    def unique_ids(self):
        """Unique IDs used in the Graph"""
        return self._adjacency.index.get_level_values("focal").unique()

    @cached_property
    def n(self):
        """Number of observations."""
        return self.unique_ids.shape[0]

    @cached_property
    def n_nodes(self):
        """Number of observations."""
        return self.unique_ids.shape[0]

    @cached_property
    def n_edges(self):
        """Number of observations."""
        return self._adjacency.shape[0] - self.isolates.shape[0]

    @cached_property
    def pct_nonzero(self):
        """Percentage of nonzero weights."""
        p = 100.0 * self.sparse.nnz / (1.0 * self.n**2)
        return p

    @cached_property
    def nonzero(self):
        """Number of nonzero weights."""
        return (self._adjacency.drop(self.isolates) > 0).sum()

    def asymmetry(self, intrinsic=True):
        r"""Asymmetry check.

        Parameters
        ----------
        intrinsic : bool, optional
            Default is ``True``. Intrinsic symmetry is defined as:

            .. math::

                w_{i,j} == w_{j,i}

            If ``intrinsic`` is ``False`` symmetry is defined as:

            .. math::

                i \in N_j \ \& \ j \in N_i

            where :math:`N_j` is the set of neighbors for :math:`j`,
            e.g., ``True`` requires equality of the weight to consider
            two links equal, ``False`` requires only a presence of a link
            with a non-zero weight.

        Returns
        -------
        pandas.Series
            A ``Series`` of ``(i,j)`` pairs of asymmetries sorted
            ascending by the focal observation (index value),
            where ``i`` is the focal and ``j`` is the neighbor.
            An empty ``Series`` is returned if no asymmetries are found.
        """
        if intrinsic:
            wd = self.sparse.transpose() - self.sparse
        else:
            transformed = self.transform("b")
            wd = transformed.sparse.transpose() - transformed.sparse

        ids = np.nonzero(wd)
        if len(ids[0]) == 0:
            return pd.Series(
                index=pd.Index([], name="focal"),
                name="neighbor",
                dtype=self._adjacency.index.dtypes["focal"],
            )
        else:
            i2id = dict(
                zip(np.arange(self.unique_ids.shape[0]), self.unique_ids, strict=True)
            )
            focal, neighbor = np.nonzero(wd)
            focal = focal.astype(self._adjacency.index.dtypes["focal"])
            neighbor = neighbor.astype(self._adjacency.index.dtypes["focal"])
            for i in i2id:
                focal[focal == i] = i2id[i]
                neighbor[neighbor == i] = i2id[i]
            ijs = pd.Series(
                neighbor, index=pd.Index(focal, name="focal"), name="neighbor"
            ).sort_index()
            return ijs

    def higher_order(self, k=2, shortest_path=True, diagonal=False, lower_order=False):
        """Contiguity weights object of order :math:`k`.

        Proper higher order neighbors are returned such that :math:`i` and :math:`j`
        are :math:`k`-order neighbors if the shortest path from :math:`i-j` is of
        length :math:`k`.

        Parameters
        ----------
        k : int, optional
            Order of contiguity. By default 2.
        shortest_path : bool, optional
            If True, :math:`i,j` and :math:`k`-order neighbors if the shortest
            path for :math:`i,j` is :math:`k`. If False, :math:`i,j` are
            `k`-order neighbors if there is a path from :math:`i,j` of length
            :math:`k`. By default True.
        diagonal : bool, optional
            If True, keep :math:`k`-order (:math:`i,j`) joins when :math:`i==j`.
            If False, remove :math:`k`-order (:math:`i,j`) joins when
            :math:`i==j`. By default False.
        lower_order : bool, optional
            If True, include lower order contiguities. If False return only weights of
            order :math:`k`. By default False.

        Returns
        -------
        Graph
            higher order weights
        """
        # TODO: remove casting to matrix once scipy implements matrix_power on array.
        # [https://github.com/scipy/scipy/pull/18544]
        binary = self.transform("B")
        sp = sparse.csr_matrix(binary.sparse)

        if lower_order:
            wk = sum(sp**x for x in range(2, k + 1))
            shortest_path = False
        else:
            wk = sp**k

        rk, ck = wk.nonzero()
        sk = set(zip(rk, ck, strict=True))

        if shortest_path:
            for j in range(1, k):
                wj = sp**j
                rj, cj = wj.nonzero()
                sj = set(zip(rj, cj, strict=True))
                sk.difference_update(sj)
        if not diagonal:
            sk = {(i, j) for i, j in sk if i != j}

        return Graph.from_sparse(
            sparse.coo_array(
                (
                    np.ones(len(sk), dtype=np.int8),
                    ([s[0] for s in sk], [s[1] for s in sk]),
                ),
                shape=sp.shape,
            ),
            ids=self.unique_ids,
        )

    def lag(self, y):
        """Spatial lag operator

        If weights are row standardized, returns the mean of each
        observation's neighbors; if not, returns the weighted sum
        of each observation's neighbors.

        Parameters
        ----------
        y : array-like
            array-like (N,) shape where N is equal to number of observations in self.

        Returns
        -------
        numpy.ndarray
            array of numeric values for the spatial lag
        """
        return _lag_spatial(self, y)

    def to_parquet(self, path, **kwargs):
        """Save Graph to a Apache Parquet

        Graph is serialized to the Apache Parquet using the underlying adjacency
        object stored as a Parquet table and custom metadata containing transformation.

        Requires pyarrow package.

        Parameters
        ----------
        path : str | pyarrow.NativeFile
            path or any stream supported by pyarrow
        **kwargs
            additional keyword arguments passed to pyarrow.parquet.write_table

        See also
        --------
        read_parquet
        """
        _to_parquet(self, path, **kwargs)

    def to_networkx(self):
        """Convert Graph to a ``networkx`` graph.

        If Graph is symmetric, returns ``nx.Graph``, otherwise returns a ``nx.DiGraph``.

        Returns
        -------
        networkx.Graph | networkx.DiGraph
            Representation of libpysal Graph as networkx graph
        """
        try:
            import networkx as nx
        except ImportError:
            raise ImportError("NetworkX is required.") from None

        graph_type = nx.Graph if self.asymmetry().empty else nx.DiGraph

        return nx.from_pandas_edgelist(
            self._adjacency.reset_index(),
            source="focal",
            target="neighbor",
            edge_attr="weight",
            create_using=graph_type,
        )

    def plot(
        self,
        gdf,
        focal=None,
        nodes=True,
        color="k",
        edge_kws=None,
        node_kws=None,
        focal_kws=None,
        ax=None,
        figsize=None,
        limit_extent=False,
    ):
        """Plot edges and nodes of the Graph

        Creates a ``maptlotlib`` plot based on the topology stored in the
        Graph and spatial location defined in ``gdf``.

        Parameters
        ----------
        gdf : geopandas.GeoDataFrame
            Geometries indexed using the same index as Graph. Geometry types other than
            points are converted to centroids encoding start and end point of Graph
            edges.
        focal : hashable | array-like[hashable] | None, optional
            ID or an array-like of IDs of focal geometries whose weights shall be
            plotted. If None, all weights from all focal geometries are plotted.
            By default None
        nodes : bool, optional
            Plot nodes as points, by default True
        color : str, optional
            The color of all objects, by default "k"
        edge_kws : dict, optional
            Keyword arguments dictionary to send to ``LineCollection``,
            which provides fine-grained control over the aesthetics
            of the edges in the plot. By default None
        node_kws : dict, optional
            Keyword arguments dictionary to send to ``ax.scatter``,
            which provides fine-grained control over the aesthetics
            of the nodes in the plot. By default None
        focal_kws : dict, optional
            Keyword arguments dictionary to send to ``ax.scatter``,
            which provides fine-grained control over the aesthetics
            of the focal nodes in the plot on top of generic ``node_kws``.
            Values of ``node_kws`` are updated from ``focal_kws``.
            Ignored if ``focal=None``. By default None
        ax : matplotlib.axes.Axes, optional
            Axis on which to plot the weights. If None, a new figure and axis are
            created. By default None
        figsize : tuple, optional
            figsize used to create a new axis. By default None
        limit_extent : bool, optional
            limit the extent of the axis to the extent of the plotted graph, by default
            False

        Returns
        -------
        matplotlib.axes.Axes
            Axis with the resulting plot

        Notes
        -----
        If you'd like to overlay the actual geometries from the
        ``geopandas.GeoDataFrame``, create an axis by plotting the ``GeoDataFrame``
        and plot the Graph on top.

            ax = gdf.plot()
            gdf_graph.plot(gdf, ax=ax)

        """
        return _plot(
            self,
            gdf,
            focal=focal,
            nodes=nodes,
            color=color,
            node_kws=node_kws,
            edge_kws=edge_kws,
            focal_kws=focal_kws,
            ax=ax,
            figsize=figsize,
            limit_extent=limit_extent,
        )

    def explore(
        self,
        gdf,
        focal=None,
        nodes=True,
        color="black",
        edge_kws=None,
        node_kws=None,
        focal_kws=None,
        m=None,
    ):
        """Plot graph as an interactive Folium Map

        Parameters
        ----------
        gdf : geopandas.GeoDataFrame
            geodataframe used to instantiate to Graph
        focal : list, optional
            subset of focal observations to plot in the map, by default None.
            If none, all relationships are plotted
        nodes : bool, optional
            whether to display observations as nodes in the map, by default True
        color : str, optional
            color applied to nodes and edges, by default "black"
        edge_kws : dict, optional
            additional keyword arguments passed to geopandas explore function
            when plotting edges, by default None
        node_kws : dict, optional
            additional keyword arguments passed to geopandas explore function
            when plotting nodes, by default None
        focal_kws : dict, optional
            additional keyword arguments passed to geopandas explore function
            when plotting focal observations, by default None. Only applicable when
            passing a subset of nodes with the `focal` argument
        m : Folilum.Map, optional
            folium map objecto to plot on top of, by default None

        Returns
        -------
        folium.Map
            folium map
        """
        return _explore_graph(
            self,
            gdf,
            focal=focal,
            nodes=nodes,
            color=color,
            edge_kws=edge_kws,
            node_kws=node_kws,
            focal_kws=focal_kws,
            m=m,
        )


def _arrange_arrays(heads, tails, weights, ids=None):
    """
    Rearrange input arrays so that observation indices
    are well-ordered with respect to the input ids. That is,
    an "early" identifier should always preceed a "later" identifier
    in the heads, but the tails should be sorted with respect
    to heads *first*, then sorted within the tails.
    """
    if ids is None:
        ids = np.unique(np.hstack((heads, tails)))
    lookup = list(ids).index
    input_df = pd.DataFrame.from_dict(
        {"focal": heads, "neighbor": tails, "weight": weights}
    )
    return (
        input_df.set_index(["focal", "neighbor"])
        .assign(
            focal_loc=input_df.focal.apply(lookup).values,
            neighbor_loc=input_df.neighbor.apply(lookup).values,
        )
        .sort_values(["focal_loc", "neighbor_loc"])
        .reset_index()
        .drop(["focal_loc", "neighbor_loc"], axis=1)
        .values.T
    )


def read_parquet(path, **kwargs):
    """Read Graph from a Apache Parquet

    Read Graph serialized using `Graph.to_parquet()` back into the `Graph` object.
    The Parquet file needs to contain adjacency table with a structure required
    by the `Graph` constructor and optional metadata with the type of transformation.

    Parameters
    ----------
    path : str | pyarrow.NativeFile | file-like object
        path or any stream supported by pyarrow
    **kwargs
        additional keyword arguments passed to pyarrow.parquet.read_table

    Returns
    -------
    Graph
        deserialized Graph
    """
    adjacency, transformation = _read_parquet(path, **kwargs)
    return Graph(adjacency, transformation)



import os
import tempfile
import warnings

import geopandas
import numpy as np
import pytest
import scipy.sparse

from ... import examples
from ...io.fileio import FileIO as psopen
from .. import util
from ..contiguity import Rook
from ..distance import KNN
from ..util import WSP2W, lat2W
from ..weights import WSP, W, _LabelEncoder


class TestW:
    def setup_method(self):
        self.w = Rook.from_shapefile(
            examples.get_path("10740.shp"), silence_warnings=True
        )

        self.neighbors = {
            0: [3, 1],
            1: [0, 4, 2],
            2: [1, 5],
            3: [0, 6, 4],
            4: [1, 3, 7, 5],
            5: [2, 4, 8],
            6: [3, 7],
            7: [4, 6, 8],
            8: [5, 7],
        }
        self.weights = {
            0: [1, 1],
            1: [1, 1, 1],
            2: [1, 1],
            3: [1, 1, 1],
            4: [1, 1, 1, 1],
            5: [1, 1, 1],
            6: [1, 1],
            7: [1, 1, 1],
            8: [1, 1],
        }

        self.w3x3 = util.lat2W(3, 3)
        self.w_islands = W({0: [1], 1: [0, 2], 2: [1], 3: []})

    def test_w(self):
        w = W(self.neighbors, self.weights, silence_warnings=True)
        assert w.pct_nonzero == 29.62962962962963

    def test___getitem__(self):
        assert self.w[0] == {1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}

    def test___init__(self):
        w = W(self.neighbors, self.weights, silence_warnings=True)
        assert w.pct_nonzero == 29.62962962962963

    def test___iter__(self):
        w = lat2W(3, 3)
        res = {}
        for i, wi in enumerate(w):
            res[i] = wi
        assert res[0] == (0, {1: 1.0, 3: 1.0})
        assert res[8] == (8, {5: 1.0, 7: 1.0})

    def test_asymmetries(self):
        w = lat2W(3, 3)
        w.transform = "r"
        result = w.asymmetry()
        assert result == [
            (0, 1),
            (0, 3),
            (1, 0),
            (1, 2),
            (1, 4),
            (2, 1),
            (2, 5),
            (3, 0),
            (3, 4),
            (3, 6),
            (4, 1),
            (4, 3),
            (4, 5),
            (4, 7),
            (5, 2),
            (5, 4),
            (5, 8),
            (6, 3),
            (6, 7),
            (7, 4),
            (7, 6),
            (7, 8),
            (8, 5),
            (8, 7),
        ]

    def test_asymmetry(self):
        w = lat2W(3, 3)
        assert w.asymmetry() == []
        w.transform = "r"
        assert not w.asymmetry() == []

    def test_asymmetry_string_index(self):
        neighbors = {
            "a": ["b", "c", "d"],
            "b": ["b", "c", "d"],
            "c": ["a", "b"],
            "d": ["a", "b"],
        }
        weights_d = {"a": [1, 1, 1], "b": [1, 1, 1], "c": [1, 1], "d": [1, 1]}
        w = W(neighbors, weights_d)
        assert w.asymmetry() == [("a", "b"), ("b", "a")]

        w.transform = "r"
        assert w.asymmetry() == [
            ("a", "b"),
            ("a", "c"),
            ("a", "d"),
            ("b", "a"),
            ("b", "c"),
            ("b", "d"),
            ("c", "a"),
            ("c", "b"),
            ("d", "a"),
            ("d", "b"),
        ]

    def test_asymmetry_mixed_index(self):
        neighbors = {
            3000: [45, 99.99, "-"],
            45: [45, 99.99, "-"],
            99.99: [3000, 45],
            "-": [3000, 45],
        }
        weights_d = {3000: [1, 1, 1], 45: [1, 1, 1], 99.99: [1, 1], "-": [1, 1]}
        w = W(neighbors, weights_d, id_order=list(neighbors.keys()))
        assert w.asymmetry() == [(3000, 45), (45, 3000)]

        w.transform = "r"
        assert w.asymmetry() == [
            (3000, 45),
            (3000, 99.99),
            (3000, "-"),
            (45, 3000),
            (45, 99.99),
            (45, "-"),
            (99.99, 3000),
            (99.99, 45),
            ("-", 3000),
            ("-", 45),
        ]

    def test_cardinalities(self):
        w = lat2W(3, 3)
        assert w.cardinalities == {0: 2, 1: 3, 2: 2, 3: 3, 4: 4, 5: 3, 6: 2, 7: 3, 8: 2}

    def test_diag_w2(self):
        np.testing.assert_array_almost_equal(
            self.w3x3.diagW2, np.array([2.0, 3.0, 2.0, 3.0, 4.0, 3.0, 2.0, 3.0, 2.0])
        )

    def test_diag_wt_w(self):
        np.testing.assert_array_almost_equal(
            self.w3x3.diagW2, np.array([2.0, 3.0, 2.0, 3.0, 4.0, 3.0, 2.0, 3.0, 2.0])
        )

    def test_diag_wt_w_ww(self):
        np.testing.assert_array_almost_equal(
            self.w3x3.diagWtW_WW,
            np.array([4.0, 6.0, 4.0, 6.0, 8.0, 6.0, 4.0, 6.0, 4.0]),
        )

    def test_full(self):
        wf = np.array(
            [
                [0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0],
            ]
        )
        ids = list(range(9))

        wf1, ids1 = self.w3x3.full()
        np.testing.assert_array_almost_equal(wf1, wf)
        assert ids1 == ids

    def test_get_transform(self):
        assert self.w3x3.transform == "O"
        self.w3x3.transform = "r"
        assert self.w3x3.transform == "R"
        self.w3x3.transform = "b"

    def test_higher_order(self):
        weights = {
            0: [1.0, 1.0, 1.0],
            1: [1.0, 1.0, 1.0],
            2: [1.0, 1.0, 1.0],
            3: [1.0, 1.0, 1.0],
            4: [1.0, 1.0, 1.0, 1.0],
            5: [1.0, 1.0, 1.0],
            6: [1.0, 1.0, 1.0],
            7: [1.0, 1.0, 1.0],
            8: [1.0, 1.0, 1.0],
        }
        neighbors = {
            0: [4, 6, 2],
            1: [3, 5, 7],
            2: [8, 0, 4],
            3: [7, 1, 5],
            4: [8, 0, 2, 6],
            5: [1, 3, 7],
            6: [4, 0, 8],
            7: [3, 1, 5],
            8: [6, 2, 4],
        }
        wneighbs = {
            k: {neighb: weights[k][i] for i, neighb in enumerate(v)}
            for k, v in list(neighbors.items())
        }
        w2 = util.higher_order(self.w3x3, 2)
        test_wneighbs = {
            k: {ne: weights[k][i] for i, ne in enumerate(v)}
            for k, v in list(w2.neighbors.items())
        }
        assert test_wneighbs == wneighbs

    def test_histogram(self):
        hist = [
            (0, 1),
            (1, 1),
            (2, 4),
            (3, 20),
            (4, 57),
            (5, 44),
            (6, 36),
            (7, 15),
            (8, 7),
            (9, 1),
            (10, 6),
            (11, 0),
            (12, 2),
            (13, 0),
            (14, 0),
            (15, 1),
        ]
        assert self.w.histogram == hist

    def test_id2i(self):
        id2i = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8}
        assert self.w3x3.id2i == id2i

    def test_id_order_set(self):
        w = W(neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"]})
        assert not w.id_order_set

    def test_islands(self):
        w = W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        assert w.islands == ["d"]
        assert self.w3x3.islands == []

    def test_max_neighbors(self):
        w = W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        assert w.max_neighbors == 2
        assert self.w3x3.max_neighbors == 4

    def test_mean_neighbors(self):
        w = util.lat2W()
        assert w.mean_neighbors == 3.2

    def test_min_neighbors(self):
        w = util.lat2W()
        assert w.min_neighbors == 2

    def test_n(self):
        w = util.lat2W()
        assert w.n == 25

    def test_neighbor_offsets(self):
        d = {
            0: [3, 1],
            1: [0, 4, 2],
            2: [1, 5],
            3: [0, 6, 4],
            4: [1, 3, 7, 5],
            5: [2, 4, 8],
            6: [3, 7],
            7: [4, 6, 8],
            8: [5, 7],
        }

        assert self.w3x3.neighbor_offsets == d

    def test_nonzero(self):
        assert self.w3x3.nonzero == 24

    def test_order(self):
        w = util.lat2W(3, 3)
        o = {
            0: [-1, 1, 2, 1, 2, 3, 2, 3, 0],
            1: [1, -1, 1, 2, 1, 2, 3, 2, 3],
            2: [2, 1, -1, 3, 2, 1, 0, 3, 2],
            3: [1, 2, 3, -1, 1, 2, 1, 2, 3],
            4: [2, 1, 2, 1, -1, 1, 2, 1, 2],
            5: [3, 2, 1, 2, 1, -1, 3, 2, 1],
            6: [2, 3, 0, 1, 2, 3, -1, 1, 2],
            7: [3, 2, 3, 2, 1, 2, 1, -1, 1],
            8: [0, 3, 2, 3, 2, 1, 2, 1, -1],
        }
        assert util.order(w) == o

    def test_pct_nonzero(self):
        assert self.w3x3.pct_nonzero == 29.62962962962963

    def test_s0(self):
        assert self.w3x3.s0 == 24.0

    def test_s1(self):
        assert self.w3x3.s1 == 48.0

    def test_s2(self):
        assert self.w3x3.s2 == 272.0

    def test_s2array(self):
        s2a = np.array(
            [[16.0], [36.0], [16.0], [36.0], [64.0], [36.0], [16.0], [36.0], [16.0]]
        )
        np.testing.assert_array_almost_equal(self.w3x3.s2array, s2a)

    def test_sd(self):
        assert self.w3x3.sd == 0.66666666666666663

    def test_set_transform(self):
        w = util.lat2W(2, 2)
        assert w.transform == "O"
        assert w.weights[0] == [1.0, 1.0]
        w.transform = "r"
        assert w.weights[0] == [0.5, 0.5]

    def test_shimbel(self):
        d = {
            0: [-1, 1, 2, 1, 2, 3, 2, 3, 4],
            1: [1, -1, 1, 2, 1, 2, 3, 2, 3],
            2: [2, 1, -1, 3, 2, 1, 4, 3, 2],
            3: [1, 2, 3, -1, 1, 2, 1, 2, 3],
            4: [2, 1, 2, 1, -1, 1, 2, 1, 2],
            5: [3, 2, 1, 2, 1, -1, 3, 2, 1],
            6: [2, 3, 4, 1, 2, 3, -1, 1, 2],
            7: [3, 2, 3, 2, 1, 2, 1, -1, 1],
            8: [4, 3, 2, 3, 2, 1, 2, 1, -1],
        }
        assert util.shimbel(self.w3x3) == d

    def test_sparse(self):
        assert self.w3x3.sparse.nnz == 24

    def test_trc_w2(self):
        assert self.w3x3.trcW2 == 24.0

    def test_trc_wt_w(self):
        assert self.w3x3.trcWtW == 24.0

    def test_trc_wt_w_ww(self):
        assert self.w3x3.trcWtW_WW == 48.0

    def test_symmetrize(self):
        symm = self.w.symmetrize()
        np.testing.assert_allclose(symm.sparse.toarray(), self.w.sparse.toarray())
        knn = KNN.from_shapefile(
            examples.get_path("baltim.shp"), k=10, silence_warnings=True
        )
        sknn = knn.symmetrize()
        assert not np.allclose(knn.sparse.toarray(), sknn.sparse.toarray())
        np.testing.assert_allclose(sknn.sparse.toarray(), sknn.sparse.toarray().T)
        knn.symmetrize(inplace=True)
        np.testing.assert_allclose(sknn.sparse.toarray(), knn.sparse.toarray())
        np.testing.assert_allclose(knn.sparse.toarray().T, knn.sparse.toarray())

    def test_connected_components(self):
        disco = {0: [1], 1: [0], 2: [3], 3: [2]}
        disco = W(disco)
        assert disco.n_components == 2

    def test_roundtrip_write(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            path = os.path.join(str(tmpdir), "tmp.gal")
            self.w.to_file(path)
            new = W.from_file(path)
        np.testing.assert_array_equal(self.w.sparse.toarray(), new.sparse.toarray())

    def test_plot(self):
        pytest.importorskip("matplotlib")
        df = geopandas.read_file(examples.get_path("10740.shp"))
        with warnings.catch_warnings(record=True) as record:
            self.w.plot(df)
        assert len(record) == 0

    def test_to_sparse(self):
        sparse = self.w_islands.to_sparse()
        np.testing.assert_array_equal(sparse.data, [1, 1, 1, 1, 0])
        np.testing.assert_array_equal(sparse.row, [0, 1, 1, 2, 3])
        np.testing.assert_array_equal(sparse.col, [1, 0, 2, 1, 3])
        sparse = self.w_islands.to_sparse("bsr")
        assert isinstance(sparse, scipy.sparse.bsr_array)
        sparse = self.w_islands.to_sparse("csr")
        assert isinstance(sparse, scipy.sparse.csr_array)
        sparse = self.w_islands.to_sparse("coo")
        assert isinstance(sparse, scipy.sparse.coo_array)
        sparse = self.w_islands.to_sparse("csc")
        assert isinstance(sparse, scipy.sparse.csc_array)
        sparse = self.w_islands.to_sparse()
        assert isinstance(sparse, scipy.sparse.coo_array)

    def test_sparse_fmt(self):
        with pytest.raises(ValueError) as exc_info:
            sparse = self.w_islands.to_sparse("dog")

    def test_from_sparse(self):
        sparse = self.w_islands.to_sparse()
        w = W.from_sparse(sparse)
        assert w.n == 4
        assert len(w.islands) == 0
        assert w.neighbors[3] == [3]


class Test_WSP_Back_To_W:
    # Test to make sure we get back to the same W functionality
    def setup_method(self):
        self.w = Rook.from_shapefile(
            examples.get_path("10740.shp"), silence_warnings=True
        )
        wsp = self.w.to_WSP()
        self.w = wsp.to_W(silence_warnings=True)

        self.neighbors = {
            0: [3, 1],
            1: [0, 4, 2],
            2: [1, 5],
            3: [0, 6, 4],
            4: [1, 3, 7, 5],
            5: [2, 4, 8],
            6: [3, 7],
            7: [4, 6, 8],
            8: [5, 7],
        }
        self.weights = {
            0: [1, 1],
            1: [1, 1, 1],
            2: [1, 1],
            3: [1, 1, 1],
            4: [1, 1, 1, 1],
            5: [1, 1, 1],
            6: [1, 1],
            7: [1, 1, 1],
            8: [1, 1],
        }

        self.w3x3 = util.lat2W(3, 3)
        w3x3 = WSP(self.w3x3.sparse, self.w3x3.id_order)
        self.w3x3 = WSP2W(w3x3)

    def test_w(self):
        w = W(self.neighbors, self.weights, silence_warnings=True)
        assert w.pct_nonzero == 29.62962962962963

    def test___getitem__(self):
        assert self.w[0] == {1: 1.0, 4: 1.0, 101: 1.0, 85: 1.0, 5: 1.0}

    def test___init__(self):
        w = W(self.neighbors, self.weights, silence_warnings=True)
        assert w.pct_nonzero == 29.62962962962963

    def test___iter__(self):
        w = util.lat2W(3, 3)
        res = {}
        for i, wi in enumerate(w):
            res[i] = wi
        assert res[0] == (0, {1: 1.0, 3: 1.0})
        assert res[8] == (8, {5: 1.0, 7: 1.0})

    def test_asymmetries(self):
        w = util.lat2W(3, 3)
        w.transform = "r"
        result = w.asymmetry()
        assert result == [
            (0, 1),
            (0, 3),
            (1, 0),
            (1, 2),
            (1, 4),
            (2, 1),
            (2, 5),
            (3, 0),
            (3, 4),
            (3, 6),
            (4, 1),
            (4, 3),
            (4, 5),
            (4, 7),
            (5, 2),
            (5, 4),
            (5, 8),
            (6, 3),
            (6, 7),
            (7, 4),
            (7, 6),
            (7, 8),
            (8, 5),
            (8, 7),
        ]

    def test_asymmetry(self):
        w = util.lat2W(3, 3)
        assert w.asymmetry() == []
        w.transform = "r"
        assert not w.asymmetry() == []

    def test_cardinalities(self):
        w = util.lat2W(3, 3)
        assert w.cardinalities == {0: 2, 1: 3, 2: 2, 3: 3, 4: 4, 5: 3, 6: 2, 7: 3, 8: 2}

    def test_diag_w2(self):
        np.testing.assert_array_almost_equal(
            self.w3x3.diagW2, np.array([2.0, 3.0, 2.0, 3.0, 4.0, 3.0, 2.0, 3.0, 2.0])
        )

    def test_diag_wt_w(self):
        np.testing.assert_array_almost_equal(
            self.w3x3.diagW2, np.array([2.0, 3.0, 2.0, 3.0, 4.0, 3.0, 2.0, 3.0, 2.0])
        )

    def test_diag_wt_w_ww(self):
        np.testing.assert_array_almost_equal(
            self.w3x3.diagWtW_WW,
            np.array([4.0, 6.0, 4.0, 6.0, 8.0, 6.0, 4.0, 6.0, 4.0]),
        )

    def test_full(self):
        wf = np.array(
            [
                [0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0],
                [1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 1.0, 0.0],
            ]
        )
        ids = list(range(9))

        wf1, ids1 = self.w3x3.full()
        np.testing.assert_array_almost_equal(wf1, wf)
        assert ids1 == ids

    def test_get_transform(self):
        assert self.w3x3.transform == "O"
        self.w3x3.transform = "r"
        assert self.w3x3.transform == "R"
        self.w3x3.transform = "b"

    def test_higher_order(self):
        weights = {
            0: [1.0, 1.0, 1.0],
            1: [1.0, 1.0, 1.0],
            2: [1.0, 1.0, 1.0],
            3: [1.0, 1.0, 1.0],
            4: [1.0, 1.0, 1.0, 1.0],
            5: [1.0, 1.0, 1.0],
            6: [1.0, 1.0, 1.0],
            7: [1.0, 1.0, 1.0],
            8: [1.0, 1.0, 1.0],
        }
        neighbors = {
            0: [4, 6, 2],
            1: [3, 5, 7],
            2: [8, 0, 4],
            3: [7, 1, 5],
            4: [8, 0, 2, 6],
            5: [1, 3, 7],
            6: [4, 0, 8],
            7: [3, 1, 5],
            8: [6, 2, 4],
        }
        wneighbs = {
            k: {neighb: weights[k][i] for i, neighb in enumerate(v)}
            for k, v in list(neighbors.items())
        }
        w2 = util.higher_order(self.w3x3, 2)
        test_wneighbs = {
            k: {ne: w2.weights[k][i] for i, ne in enumerate(v)}
            for k, v in list(w2.neighbors.items())
        }
        assert test_wneighbs == wneighbs

    def test_histogram(self):
        hist = [
            (0, 1),
            (1, 1),
            (2, 4),
            (3, 20),
            (4, 57),
            (5, 44),
            (6, 36),
            (7, 15),
            (8, 7),
            (9, 1),
            (10, 6),
            (11, 0),
            (12, 2),
            (13, 0),
            (14, 0),
            (15, 1),
        ]
        assert self.w.histogram == hist

    def test_id2i(self):
        id2i = {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6, 7: 7, 8: 8}
        assert self.w3x3.id2i == id2i

    def test_id_order_set(self):
        w = W(neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"]})
        assert not w.id_order_set

    def test_islands(self):
        w = W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        assert w.islands == ["d"]
        assert self.w3x3.islands == []

    def test_max_neighbors(self):
        w = W(
            neighbors={"a": ["b"], "b": ["a", "c"], "c": ["b"], "d": []},
            silence_warnings=True,
        )
        assert w.max_neighbors == 2
        assert self.w3x3.max_neighbors == 4

    def test_mean_neighbors(self):
        w = util.lat2W()
        assert w.mean_neighbors == 3.2

    def test_min_neighbors(self):
        w = util.lat2W()
        assert w.min_neighbors == 2

    def test_n(self):
        w = util.lat2W()
        assert w.n == 25

    def test_nonzero(self):
        assert self.w3x3.nonzero == 24

    def test_order(self):
        w = util.lat2W(3, 3)
        o = {
            0: [-1, 1, 2, 1, 2, 3, 2, 3, 0],
            1: [1, -1, 1, 2, 1, 2, 3, 2, 3],
            2: [2, 1, -1, 3, 2, 1, 0, 3, 2],
            3: [1, 2, 3, -1, 1, 2, 1, 2, 3],
            4: [2, 1, 2, 1, -1, 1, 2, 1, 2],
            5: [3, 2, 1, 2, 1, -1, 3, 2, 1],
            6: [2, 3, 0, 1, 2, 3, -1, 1, 2],
            7: [3, 2, 3, 2, 1, 2, 1, -1, 1],
            8: [0, 3, 2, 3, 2, 1, 2, 1, -1],
        }
        assert util.order(w) == o

    def test_pct_nonzero(self):
        assert self.w3x3.pct_nonzero == 29.62962962962963

    def test_s0(self):
        assert self.w3x3.s0 == 24.0

    def test_s1(self):
        assert self.w3x3.s1 == 48.0

    def test_s2(self):
        assert self.w3x3.s2 == 272.0

    def test_s2array(self):
        s2a = np.array(
            [[16.0], [36.0], [16.0], [36.0], [64.0], [36.0], [16.0], [36.0], [16.0]]
        )
        np.testing.assert_array_almost_equal(self.w3x3.s2array, s2a)

    def test_sd(self):
        assert self.w3x3.sd == 0.66666666666666663

    def test_set_transform(self):
        w = util.lat2W(2, 2)
        assert w.transform == "O"
        assert w.weights[0] == [1.0, 1.0]
        w.transform = "r"
        assert w.weights[0] == [0.5, 0.5]

    def test_shimbel(self):
        d = {
            0: [-1, 1, 2, 1, 2, 3, 2, 3, 4],
            1: [1, -1, 1, 2, 1, 2, 3, 2, 3],
            2: [2, 1, -1, 3, 2, 1, 4, 3, 2],
            3: [1, 2, 3, -1, 1, 2, 1, 2, 3],
            4: [2, 1, 2, 1, -1, 1, 2, 1, 2],
            5: [3, 2, 1, 2, 1, -1, 3, 2, 1],
            6: [2, 3, 4, 1, 2, 3, -1, 1, 2],
            7: [3, 2, 3, 2, 1, 2, 1, -1, 1],
            8: [4, 3, 2, 3, 2, 1, 2, 1, -1],
        }
        assert util.shimbel(self.w3x3) == d

    def test_sparse(self):
        assert self.w3x3.sparse.nnz == 24

    def test_trc_w2(self):
        assert self.w3x3.trcW2 == 24.0

    def test_trc_wt_w(self):
        assert self.w3x3.trcWtW == 24.0

    def test_trc_wt_w_ww(self):
        assert self.w3x3.trcWtW_WW == 48.0


class TestWSP:
    def setup_method(self):
        self.w = psopen(examples.get_path("sids2.gal")).read()
        self.wsp = WSP(self.w.sparse, self.w.id_order)
        w3x3 = util.lat2W(3, 3)
        self.w3x3 = WSP(w3x3.sparse)

    def test_wsp(self):
        assert self.w.id_order == self.wsp.id_order
        assert self.w.n == self.wsp.n
        np.testing.assert_array_equal(
            self.w.sparse.todense(), self.wsp.sparse.todense()
        )

    def test_diag_wt_w_ww(self):
        np.testing.assert_array_almost_equal(
            self.w3x3.diagWtW_WW,
            np.array([4.0, 6.0, 4.0, 6.0, 8.0, 6.0, 4.0, 6.0, 4.0]),
        )

    def test_trc_wt_w_ww(self):
        assert self.w3x3.trcWtW_WW == 48.0

    def test_s0(self):
        assert self.w3x3.s0 == 24.0

    def test_from_wsp(self):
        w = W.from_WSP(self.wsp)
        assert w.n == 100
        assert w.pct_nonzero == 4.62

    def test_label_encoder(self):
        le = _LabelEncoder()
        le.fit(["NY", "CA", "NY", "CA", "TX", "TX"])
        np.testing.assert_equal(le.classes_, np.array(["CA", "NY", "TX"]))
        np.testing.assert_equal(
            le.transform(["NY", "CA", "NY", "CA", "TX", "TX"]),
            np.array([1, 0, 1, 0, 2, 2]),
        )



from .. import tables

__author__ = "Charles R Schmidt <schmidtc@gmail.com>"
__all__ = ["GeoDaTxtReader"]

from typing import Union


class GeoDaTxtReader(tables.DataTable):
    """GeoDa Text File Export Format.
    
    Examples
    --------
    
    >>> import libpysal
    >>> f = libpysal.io.open(libpysal.examples.get_path('stl_hom.txt'),'r')
    >>> f.header
    ['FIPSNO', 'HR8488', 'HR8893', 'HC8488']
    
    >>> len(f)
    78
    
    >>> f.dat[0]
    ['17107', '1.290722', '1.624458', '2']
    
    >>> f.dat[-1]
    ['29223', '0', '8.451537', '0']
    
    >>> f._spec
    [int, float, float, int]

    """

    __doc__ = tables.DataTable.__doc__

    FORMATS = ["geoda_txt"]
    MODES = ["r"]

    def __init__(self, *args, **kwargs):

        tables.DataTable.__init__(self, *args, **kwargs)
        self.__idx = {}
        self.__len = None
        self.pos = 0
        self._open()

    def _open(self):
        """
        
        Raises
        ------
        TypeError
            Raised when the input 'geoda_txt' is not valid.
        
        """

        if self.mode == "r":

            self.fileObj = open(self.dataPath, "r")
            n, k = self.fileObj.readline().strip().split(",")
            n, k = int(n), int(k)
            header = self.fileObj.readline().strip().split(",")
            self.header = [f.replace('"', "") for f in header]

            try:
                assert len(self.header) == k
            except AssertionError:
                raise TypeError("This is not a valid 'geoda_txt' file.")

            dat = self.fileObj.readlines()

            self.dat = [line.strip().split(",") for line in dat]
            self._spec = self._determineSpec(self.dat)
            self.__len = len(dat)

    def __len__(self) -> int:
        return self.__len

    def _read(self) -> Union[list, None]:
        if self.pos < len(self):
            row = self.dat[self.pos]
            self.pos += 1
            return row
        else:
            return None

    def close(self):
        self.fileObj.close()
        tables.DataTable.close(self)

    @staticmethod
    def _determineSpec(data) -> list:
        cols = len(data[0])
        spec = []

        for j in range(cols):
            isInt = True
            isFloat = True

            for row in data:
                val = row[j]
                if not val.strip().replace("-", "").replace(".", "").isdigit():
                    isInt = False
                    isFloat = False
                    break
                else:
                    if isInt and "." in val:
                        isInt = False
            if isInt:
                spec.append(int)
            elif isFloat:
                spec.append(float)
            else:
                spec.append(str)

        return spec



import geodatasets
import geopandas
import pandas as pd
import pytest

from libpysal.graph.base import Graph


class TestSetOps:
    def setup_method(self):
        self.grocs = geopandas.read_file(geodatasets.get_path("geoda groceries"))[
            ["OBJECTID", "geometry"]
        ].explode(ignore_index=True)

        self.distance500 = Graph.build_distance_band(self.grocs, 500)
        self.distance2500 = Graph.build_distance_band(self.grocs, 2500)
        self.knn3 = Graph.build_knn(self.grocs, 3)

        self.distance2500_id = Graph.build_distance_band(
            self.grocs.set_index("OBJECTID"), 2500
        )

    def test_intersects(self):
        assert self.distance2500.intersects(self.knn3)
        assert self.knn3.intersects(self.distance2500)
        assert not self.knn3.intersects(self.distance2500_id)
        assert not self.distance2500.intersects(self.distance2500_id)

    def test_intersection(self):
        result = self.distance2500.intersection(self.knn3)
        assert len(result) == 115
        assert result._adjacency.shape[0] == 185
        pd.testing.assert_index_equal(result.unique_ids, self.distance2500.unique_ids)

    def test_symmetric_difference(self):
        result = self.distance2500.symmetric_difference(self.knn3)
        assert len(result) == 334
        assert result._adjacency.shape[0] == 340
        pd.testing.assert_index_equal(result.unique_ids, self.distance2500.unique_ids)

        with pytest.raises(ValueError, match="Cannot do symmetric difference"):
            self.distance2500_id.symmetric_difference(self.knn3)

    def test_union(self):
        result = self.distance2500.union(self.knn3)
        assert len(result) == 449
        assert result._adjacency.shape[0] == 449
        pd.testing.assert_index_equal(result.unique_ids, self.distance2500.unique_ids)

        with pytest.raises(ValueError, match="Cannot do union"):
            self.distance2500_id.union(self.knn3)

    def test_difference(self):
        result = self.distance2500.difference(self.knn3)
        assert len(result) == 5
        assert result._adjacency.shape[0] == 148
        pd.testing.assert_index_equal(result.unique_ids, self.distance2500.unique_ids)

        result = self.knn3.difference(self.distance2500)
        assert len(result) == 329
        assert result._adjacency.shape[0] == 340
        pd.testing.assert_index_equal(result.unique_ids, self.knn3.unique_ids)

    def test_issubgraph(self):
        assert self.distance500.issubgraph(self.distance2500)
        assert not self.distance2500.issubgraph(self.distance500)
        assert not self.knn3.issubgraph(self.distance2500)
        assert self.knn3.issubgraph(self.knn3)

    def test_equals(self):
        assert self.distance2500.equals(self.distance2500.copy())
        assert not self.distance2500.equals(self.distance2500.transform("r"))

    def test_isomorphic(self):
        pytest.importorskip("networkx")
        assert self.distance2500.isomorphic(self.distance2500_id)

    def test___le__(self):
        assert self.distance500 <= self.distance2500
        assert not self.knn3 <= self.distance2500
        assert self.distance2500 <= self.distance2500

    def test___ge__(self):
        assert self.distance2500 >= self.distance500
        assert not self.knn3 >= self.distance2500
        assert self.distance2500 >= self.distance2500

    def test___lt__(self):
        assert self.distance500 < self.distance2500
        assert not self.knn3 < self.distance2500
        assert not self.distance2500 < self.distance2500

    def test___gt__(self):
        assert self.distance2500 > self.distance500
        assert not self.knn3 > self.distance2500
        assert not self.distance2500 > self.distance2500

    def test___eq__(self):
        assert self.distance2500 == self.distance2500.copy()
        assert not self.distance2500 == self.distance2500_id  # noqa SIM201

    def test___ne__(self):
        assert not self.distance2500 != self.distance2500.copy()  # noqa SIM202
        assert self.distance2500 != self.distance2500_id

    def test___and__(self):
        result = self.distance2500 & self.knn3
        assert len(result) == 115
        assert result._adjacency.shape[0] == 185
        pd.testing.assert_index_equal(result.unique_ids, self.distance2500.unique_ids)

    def test___or__(self):
        result = self.distance2500 | self.knn3
        assert len(result) == 449
        assert result._adjacency.shape[0] == 449
        pd.testing.assert_index_equal(result.unique_ids, self.distance2500.unique_ids)

        with pytest.raises(ValueError, match="Cannot do union"):
            self.distance2500_id | self.knn3

    def test___xor__(self):
        result = self.distance2500 ^ self.knn3
        assert len(result) == 334
        assert result._adjacency.shape[0] == 340
        pd.testing.assert_index_equal(result.unique_ids, self.distance2500.unique_ids)

        with pytest.raises(ValueError, match="Cannot do symmetric difference"):
            self.distance2500_id ^ self.knn3

    def test___iand__(self):
        with pytest.raises(TypeError, match="Graphs are immutable."):
            self.distance2500 &= self.knn3

    def test___ior__(self):
        with pytest.raises(TypeError, match="Graphs are immutable."):
            self.distance2500 |= self.knn3

    def test___len__(self):
        assert len(self.distance2500) == 120
        assert len(self.distance500) == 8
        assert len(self.knn3) == len(self.grocs) * 3
